package org.xquark.mapping;

import java.sql.Connection;

public class TriggerEvent {
	
	public static final int INSERT = 0;
	public static final int UPDATE = 1;
	public static final int DELETE = 2;
	public static final String[] actionStrings = {"insert", "update", "delete"}; 

	private Connection connection;
	private String namespace;
	private String mode;
	private String table; 
	private String[] colnames;
	private Object[] colvalues;
	private int action = -1;
	/**
	 * Constructor for TriggerEvent
	 * @param context the storage context that provides access to storage information such as the JDBC connection
	 * @param table the name of the table in which the tuple has been inserted
	 * @param colnames the names of the tuple columns that have been selected for the event
	 * @param colvalues the values of the tuple columns that have been selected for the event
	 */
	public TriggerEvent(StorageContext context, String table, String[] colnames, Object[] colvalues, int action) {
		this.connection = context.getConnection();
		this.namespace = context.getMappingNamespace();
		this.mode = context.getMappingMode();
		this.table = table;
		this.colnames = colnames;
		this.colvalues = colvalues;
		this.action = action;
	}
	
	/**
	 * Returns the JDBC connection
	 * @return the JDBC connection
	 */
	public Connection getStorageContext() {
		return connection;
	}

    /**
     * Returns the target namespace of the active mapping.
     * @return the target namespace as a String
     */
	public String getMappingNamespace() {
		return namespace;
	}
	
    /**
     * Returns the current mappping mode.
     * @return the mapping mode as a String
     */
	public String getMappingMode() {
		return mode;
	}
	
	/**
	 * Returns the table name for the event
	 * @return the table name
	 */
	public String getTableName() {
		return table;
	}
	
	/**
	 * Returns the number of columns selected for the event
	 * @return the number of columns
	 */
	public int getColumnCount() {
		return colnames.length;
	}
	
	/**
	 * Returns the array of column names
	 * @return the array of column names
	 */
	public String[] getColumnNames() {
		return colnames;
	}

	/**
	 * Returns the array of column values
	 * @return the array of column values
	 */
	public Object[] getColumnValues() {
		return colvalues;
	}
	
	/**
	 * Returns the name of the column at the given index
	 * @param index the index of the column, starting at 0
	 * @return the name of the column
	 */
	public String getColumnName(int index) {
		return colnames[index];
	}

	/**
	 * Returns the value of the column at the given index
	 * @param index the index of the column, starting at 0
	 * @return the value of the column
	 */
	public Object getColumnValue(int index) {
		return colvalues[index];
	}
	
	/**
	 * Returns the value of the column for the given column name
	 * @param colname the name of the column
	 * @return the value of the column
	 */
	public Object getColumnValue(String colname) {
		for (int i = 0; i < colnames.length; i++) {
			if (colnames[i].equals(colname))
				return colvalues[i];
		}
		return null;
	}
	
	/**
	 * Returns the action linked to this event
	 * @return one of TriggerEvent.INSERT, TriggerEvent.DELETE, TriggerEvent.UPDATE
	 */
	public int getAction() {
		return action;
	}
	
	public String toString() {
		StringBuffer buf = new StringBuffer();
		buf.append("[TriggerEvent: ");
		buf.append(actionStrings[action]);
		buf.append(" {");
		buf.append(getMappingNamespace());
		buf.append("}");
		buf.append(getMappingMode());
		buf.append(" table ");
		buf.append(getTableName());
		buf.append("(");
		for (int i = 0; i < getColumnCount(); i++) {
			if (i > 0)
				buf.append(", ");
			buf.append(getColumnName(i));
			buf.append("=");
			buf.append(getColumnValue(i));
		}
		buf.append(") ]");
		return buf.toString();
	}

}
