/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2003 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.datastructure;

import java.math.BigDecimal;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Stack;

import org.w3c.dom.*;
import org.xquark.schema.*;
import org.xquark.schema.datatypes.DateTime;
import org.xquark.schema.datatypes.Duration;
import org.xquark.schema.validation.ValidationContextProvider;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xpath.datamodel.*;
import org.xquark.xquery.EvaluatorException;
import org.xquark.xquery.metadata.DynamicContext;
import org.xquark.xquery.parser.*;
import org.xquark.xquery.parser.Element;
import org.xquark.xquery.parser.primitivefunctions.fnfunctions.*;
import org.xquark.xquery.parser.primitivefunctions.xsfunctions.*;
import org.xquark.xquery.parser.primitivefunctions.xsfunctions.FunctionSTRING;
import org.xquark.xquery.parser.util.Constants;
import org.xquark.xquery.xdbc.XDBCResultSetInterface;

public class EvaluationVisitor extends DefaultParserVisitor {
    // **********************************************************************
    // * VERSIONING
    // **********************************************************************
    private static final String RCSRevision = "$Revision: 1.16 $";

    private static final String RCSName = "$Name:  $";

    // **********************************************************************
    // * CLASS VARIABLES
    // **********************************************************************
    private ArrayList tuplelist = null;

    private Tuple tuple = null;

    private boolean resverdict = false;

    private ArrayList resnodes = null;

    private Comparable resvalue = null;

    protected static TypedDocumentImpl docimpl = null;
    static {
        docimpl = new TypedDocumentImpl();
    }

    private SchemaManager schemamanager = null;

    private XDBCResultSetInterface resultSet = null;

    private DynamicContext dc = null;

    private ValidationContextProvider context = null;

    public static final Integer UNKNOWN_TYPE = new Integer(-1);

    public static final Integer BOOLEAN_TYPE = new Integer(0);

    public static final Integer VALUE_TYPE = new Integer(1);

    public static final Integer NODE_TYPE = new Integer(2);

    private Stack returnType = new Stack();

    // only for incremental aggregate
    private int aggKind = -1;

    private final int COUNT = 0;

    private final int AVG = 1;

    private final int SUM = 2;

    private final int MIN = 3;

    private final int MAX = 4;

    private boolean generate = false;

    private boolean incrementalAggregate = false;

    /**
     * 
     */
    public EvaluationVisitor(SchemaManager schemamanager, DynamicContext dc) {
        this.schemamanager = schemamanager;
        this.dc = dc;
    }

    /**
     * 
     */
    public EvaluationVisitor(SchemaManager schemamanager, DynamicContext dc, Tuple tuple) {
        this(schemamanager, dc);
        this.tuple = tuple;
    }

    /**
     * 
     */
    public EvaluationVisitor(SchemaManager schemamanager, DynamicContext dc, ArrayList tuplelist) {
        this(schemamanager, dc);
        this.tuplelist = tuplelist;
    }

    /**
     * 
     */
    public EvaluationVisitor(SchemaManager schemamanager, DynamicContext dc, XDBCResultSetInterface resultSet, ValidationContextProvider context) {
        this(schemamanager, dc, resultSet);
        this.context = context;
    }

    /**
     * 
     */
    public EvaluationVisitor(SchemaManager schemamanager, DynamicContext dc, XDBCResultSetInterface resultSet) {
        this(schemamanager, dc);
        this.resultSet = resultSet;
    }

    public void setReturnType(Integer setType) {
        if (!returnType.isEmpty())
            this.returnType.pop();
        returnType.push(setType);
    }

    // only used for aggregate function on buffer tuples
    public void setCurrentTuple(Tuple tuple) {
        this.tuple = tuple;
    }

    public Tuple getAggregateResult() {
        Tuple tuple = new Tuple(1);
        TypedValue text = null;
        switch (aggKind) {
        case AVG:
            if (isLongSum)
                text = docimpl.createTypedValue(new BigDecimal(longSum / count), this.schemamanager.getDecimalType());
            else
                text = docimpl.createTypedValue(new Double(doubleSum / count), this.schemamanager.getDoubleType());
            break;
        case SUM:
            if (isLongSum)
                text = docimpl.createTypedValue(new Long(longSum), this.schemamanager.getIntegerType());
            else
                text = docimpl.createTypedValue(new Double(doubleSum), this.schemamanager.getDoubleType());
            break;
        case MAX:
        case MIN:
            text = docimpl.createTypedValue(comp, (SimpleType) result.getType());
            break;
        case COUNT:
            text = docimpl.createTypedValue(new Long(count), this.schemamanager.getIntegerType());
            break;
        }
        tuple.addNodeAtIndex(0, text);
        clear();
        return tuple;
    }

    private void clear() {
        resnodes = null;
        resvalue = null;
        resverdict = false;
        // for aggregate functions
        this.isLongSum = false;
        this.doubleSum = 0;
        this.longSum = 0;
        this.count = 0;
        this.comp = null;
    }

    private void setOutput(boolean verdict) throws EvaluatorException {
        this.resnodes = null;
        this.resvalue = null;
        this.resverdict = false;
        if (!returnType.isEmpty()) {
            Integer retType = (Integer) this.returnType.peek();
            if (retType == VALUE_TYPE) {
                resvalue = Boolean.toString(verdict);
                return;
                // throw new EvaluatorException("Impossible to convert boolean value to Comparable");
            } else if (retType == NODE_TYPE) {
                ArrayList nodes = new ArrayList(1);
                TypedValue text = docimpl.createTypedValue(Boolean.toString(verdict), this.schemamanager.getStringType());
                nodes.add(text);
                resnodes = nodes;
                return;
                // throw new EvaluatorException("Impossible to convert boolean value to Comparable");
            }
        }
        // default, UNKNOWN_TYPE, BOOLEAN_TYPE
        this.resverdict = verdict;
    }

    private void setOutput(Comparable value) throws EvaluatorException {
        setOutput(value, null);
    }

    private void setOutput(Comparable value, SimpleType simpleType) throws EvaluatorException {
        this.resnodes = null;
        this.resvalue = null;
        resverdict = false;
        if (!returnType.isEmpty()) {
            Integer retType = (Integer) this.returnType.peek();
            if (retType == BOOLEAN_TYPE) {
                if (value instanceof Number) {
                    this.resverdict = (value != null && value.compareTo(new Integer(0)) != 0);
                } else {
                    this.resverdict = (value != null);
                }
                return;
            } else if (retType == NODE_TYPE) {
                ArrayList nodes = null;
                if (value != null) {
                    nodes = new ArrayList(1);
                    nodes.add(docimpl.createTypedValue(value, simpleType));
                }
                this.resnodes = nodes;
                return;
            }
        }
        // default, UNKNOWN_TYPE, VALUE_TYPE
        this.resvalue = value;
    }

    private void setOutput(ArrayList nodes) throws EvaluatorException {
        this.resnodes = null;
        this.resvalue = null;
        resverdict = false;
        if (!returnType.isEmpty()) {
            Integer retType = (Integer) this.returnType.peek();
            if (retType == BOOLEAN_TYPE) {
                this.resverdict = (nodes != null && !nodes.isEmpty());
                return;
            } else if (retType == VALUE_TYPE) {
                resvalue = null;
                if (nodes != null && nodes.size() == 1 && nodes.get(0) instanceof TypedValue) {
                    TypedValue value = (TypedValue) nodes.get(0);
                    Object obj = value.getTypedValue();
                    if (obj instanceof Comparable)
                        resvalue = (Comparable) obj;
                    else
                        throw new EvaluatorException("Could not cast into Comparable");
                } else
                    throw new EvaluatorException("Could not single out multivalue");
                return;
            }
        }
        // default, UNKNOWN_TYPE, NODE_TYPE
        this.resnodes = nodes;
    }

    // private void setResultNodes(ArrayList nodes) {
    // this.resnodes = nodes;
    // this.resvalue = null;
    // this.resverdict = false;
    // }

    // private void setResultValue(Comparable comp) {
    // this.resnodes = null;
    // this.resvalue = comp;
    // resverdict = false;
    // }

    // private void setResultVerdict(boolean bool) {
    // resnodes = null;
    // resvalue = null;
    // verdict = bool;
    // }

    // public void setSchemaManager(SchemaManager schemamanager) {
    // this.schemamanager = schemamanager;
    // }
    //
    // public void setDynamicContext(DynamicContext dc) {
    // this.dc = dc;
    // }

    public void reset() {
        clear();
        this.resultSet = null;
        this.tuple = null;
        this.tuplelist = null;
        returnType.clear();
    }

    public void reset(XDBCResultSetInterface resultSet) {
        clear();
        this.resultSet = resultSet;
        returnType.clear();
    }

    public void reset(Tuple tuple) {
        clear();
        this.tuple = tuple;
        this.tuplelist = null;
        returnType.clear();
    }

    public void reset(ArrayList tuplelist) {
        clear();
        if (tuplelist.size() > 1) {
            this.tuple = null;
            this.tuplelist = tuplelist;
        } else if (tuplelist.size() == 1) {
            this.tuple = (Tuple) tuplelist.get(0);
            this.tuplelist = null;
        } else {
            this.tuple = null;
            this.tuplelist = null;
        }
        returnType.clear();
    }

    public boolean getVerdict() {
        return resverdict;
    }

    public ArrayList getResNodes() {
        return resnodes;
    }

    public Comparable getResValue() {
        // if (resvalue != null)
        // return resvalue;
        // if (resnodes != null && resnodes.size() == 1) {
        // resvalue = (Comparable) ((TypedNode)
        // resnodes.get(0)).getTypedValue();
        // }
        return resvalue;
    }

    public void doGenerate(boolean generate) {
        this.generate = generate;
    }

    public void doIncrementalAggregate(boolean incrementalAggregate) {
        this.incrementalAggregate = incrementalAggregate;
    }

    public void visit(AttributeValuePair arg) throws XQueryException {
        if (arg.isXmlns()) {
            setOutput((ArrayList) null);
            return;
        }
        if (!(arg.getExpression1() instanceof QName)) {
            throw new XQueryException("EvaluationVisitor : name of AttributeValuePair is not a QName : " + arg.getExpression1());
        }
        QName attName = (QName) arg.getExpression1();
        Object attValue = null;
        if (!(arg.getExpression2() instanceof Value)) {
            returnType.push(VALUE_TYPE);
            arg.getExpression2().accept(this);
            returnType.pop();
            if (resvalue == null) {
                throw new XQueryException("EvaluationVisitor : value of AttributeValuePair is null : " + arg.getExpression2());
            }
            attValue = resvalue;
        } else
            attValue = ((Value) arg.getExpression2()).getObjectValue();
        if (attValue != null) {
            TypedAttributeImpl att = new TypedAttributeImpl(docimpl, ((QName) attName).getNameSpace(), ((QName) attName).getLocalName());
            att.setNodeValue(attValue.toString());
            att.setTypedValue(attValue);
            ArrayList tmplist = new ArrayList(1);
            tmplist.add(att);
            setOutput(tmplist);
            return;
        }
        setOutput((ArrayList) null);
    }

    public void visit(BinOpANDExpression arg) throws XQueryException {
        boolean tmpVerdict = false;
        if (tuple != null) {
            XQueryExpression exp1 = arg.getExpression1();
            EvaluationVisitor visitor = new EvaluationVisitor(this.schemamanager, this.dc, tuple);
            visitor.setReturnType(EvaluationVisitor.BOOLEAN_TYPE);
            exp1.accept(visitor);
            tmpVerdict = visitor.getVerdict();
            if (tmpVerdict == false) {
                setOutput(false);
                return;
            }
            XQueryExpression exp2 = arg.getExpression2();
            visitor.reset(tuple);
            visitor.setReturnType(EvaluationVisitor.BOOLEAN_TYPE);
            exp2.accept(visitor);
            tmpVerdict &= visitor.getVerdict();
        } else if (tuplelist != null) {
            XQueryExpression exp1 = arg.getExpression1();
            EvaluationVisitor visitor = new EvaluationVisitor(this.schemamanager, this.dc, tuplelist);
            visitor.setReturnType(EvaluationVisitor.BOOLEAN_TYPE);
            exp1.accept(visitor);
            tmpVerdict = visitor.getVerdict();
            if (tmpVerdict == false) {
                setOutput(false);
                return;
            }
            XQueryExpression exp2 = arg.getExpression2();
            visitor.reset(tuplelist);
            visitor.setReturnType(EvaluationVisitor.BOOLEAN_TYPE);
            exp2.accept(visitor);
            tmpVerdict &= visitor.getVerdict();
        } else if (resultSet != null) {
            XQueryExpression exp1 = arg.getExpression1();
            EvaluationVisitor visitor = new EvaluationVisitor(this.schemamanager, this.dc, resultSet);
            visitor.setReturnType(EvaluationVisitor.BOOLEAN_TYPE);
            exp1.accept(visitor);
            tmpVerdict = visitor.getVerdict();
            if (!tmpVerdict) {
                setOutput(false);
                return;
            }
            XQueryExpression exp2 = arg.getExpression2();
            visitor.reset(resultSet);
            visitor.setReturnType(EvaluationVisitor.BOOLEAN_TYPE);
            exp2.accept(visitor);
            tmpVerdict &= visitor.getVerdict();
        } else {
            throw new XQueryException("BinOpANDExpression : no data available");
            // XQueryExpression exp1 = arg.getExpression1();
            // EvaluationVisitor visitor = new
            // EvaluationVisitor(this.schemamanager);
            // visitor.setReturnType(EvaluationVisitor.BOOLEAN_TYPE);
            // exp1.accept(visitor);
            // verdict = visitor.getVerdict();
            // if (verdict == false) {
            // setOutput(false);
            // return;
            // }
            // XQueryExpression exp2 = arg.getExpression2();
            // //visitor = new EvaluationVisitor(this.schemamanager);
            // visitor.reset();
            // visitor.setReturnType(EvaluationVisitor.BOOLEAN_TYPE);
            // exp2.accept(visitor);
            // verdict &= visitor.getVerdict();
        }
        setOutput(tmpVerdict);
    }

    public void visit(BinOpORExpression arg) throws XQueryException {
        boolean tmpVerdict = false;
        if (tuple != null) {
            XQueryExpression exp1 = arg.getExpression1();
            EvaluationVisitor visitor = new EvaluationVisitor(this.schemamanager, this.dc, tuple);
            visitor.setReturnType(EvaluationVisitor.BOOLEAN_TYPE);
            exp1.accept(visitor);
            tmpVerdict = visitor.getVerdict();
            XQueryExpression exp2 = arg.getExpression2();
            visitor.reset(tuple);
            visitor.setReturnType(EvaluationVisitor.BOOLEAN_TYPE);
            exp2.accept(visitor);
            tmpVerdict = tmpVerdict || visitor.getVerdict();
        } else if (tuplelist != null) {
            XQueryExpression exp1 = arg.getExpression1();
            EvaluationVisitor visitor = new EvaluationVisitor(this.schemamanager, this.dc, tuplelist);
            visitor.setReturnType(EvaluationVisitor.BOOLEAN_TYPE);
            exp1.accept(visitor);
            tmpVerdict = visitor.getVerdict();
            XQueryExpression exp2 = arg.getExpression2();
            visitor.reset(tuplelist);
            visitor.setReturnType(EvaluationVisitor.BOOLEAN_TYPE);
            exp2.accept(visitor);
            tmpVerdict = tmpVerdict || visitor.getVerdict();
        } else if (resultSet != null) {
            XQueryExpression exp1 = arg.getExpression1();
            EvaluationVisitor visitor = new EvaluationVisitor(this.schemamanager, this.dc, resultSet);
            visitor.setReturnType(EvaluationVisitor.BOOLEAN_TYPE);
            exp1.accept(visitor);
            tmpVerdict = visitor.getVerdict();
            if (tmpVerdict) {
                setOutput(true);
                return;
            }
            XQueryExpression exp2 = arg.getExpression2();
            visitor.reset(resultSet);
            visitor.setReturnType(EvaluationVisitor.BOOLEAN_TYPE);
            exp2.accept(visitor);
            tmpVerdict = tmpVerdict || visitor.getVerdict();
        } else {
            throw new XQueryException("BinOpORExpression : no data available");
            // XQueryExpression exp1 = arg.getExpression1();
            // EvaluationVisitor visitor = new
            // EvaluationVisitor(this.schemamanager);
            // exp1.accept(visitor);
            // verdict = visitor.getVerdict();
            // XQueryExpression exp2 = arg.getExpression2();
            // //visitor = new EvaluationVisitor(this.schemamanager);
            // visitor.reset();
            // exp2.accept(visitor);
            // verdict = verdict || visitor.getVerdict();
        }
        setOutput(tmpVerdict);
    }

    public void visit(Element arg) throws XQueryException {
        XQueryExpression startTag = arg.getStartTag();
        if (!(startTag instanceof QName))
            throw new XQueryException("Tag of Element expression is not a QName : " + startTag);
        Integer rettype = (Integer) returnType.peek();
        if (rettype == BOOLEAN_TYPE) {
            setOutput(true);
            return;
        } else if (rettype == UNKNOWN_TYPE) {
            rettype = NODE_TYPE;
        }
        QName startTagQName = null;
        TypedElement element = null;
        ArrayList tmplist = null;
        StringBuffer buf = null;

        if (rettype == NODE_TYPE) {
            startTagQName = (QName) startTag;
            element = new TypedElementImpl(docimpl, startTagQName.getNameSpace(), startTagQName.getLocalName());
            tmplist = arg.getAttributes();
            if (tmplist != null && !tmplist.isEmpty()) {
                for (int i = 0; i < tmplist.size(); i++) {
                    AttributeValuePair attValPair = (AttributeValuePair) tmplist.get(i);
                    attValPair.accept(this);
                    if (resnodes != null) {
                        TypedAttribute att = (TypedAttribute) resnodes.get(0);
                        element.setAttributeNodeNS((Attr) att);
                    }
                }
            }
        } else
            buf = new StringBuffer();
        tmplist = arg.getSubExpressions();
        if (tmplist != null && !tmplist.isEmpty()) {
            for (int i = 0; i < tmplist.size(); i++) {
                XQueryExpression subExpr = (XQueryExpression) tmplist.get(i);
                subExpr.accept(this);
                if (rettype == NODE_TYPE && resnodes != null) {
                    for (int j = 0; j < resnodes.size(); j++) {
                        TypedNode resnode = (TypedNode) resnodes.get(j);
                        if (docimpl != resnode.getOwnerDocument())
                            resnode = (TypedNode) element.getOwnerDocument().importNode(resnode, true);
                        if (resnode instanceof TypedAttributeImpl) {
                            element.setAttributeNodeNS((Attr) resnode);
                        } else if (resnode instanceof TypedElementImpl) {
                            element.appendChild((Node) resnode);
                        } else if (resnode instanceof TypedValueImpl) {
                            element.appendChild((Node) resnode);
                        } else {
                            throw new XQueryException("Could not add child : " + resnode + " to parent.");
                            // do nothing for now, warning?
                        }
                    }
                } else if (rettype == VALUE_TYPE && resvalue != null) {
                    buf.append(resvalue.toString());
                }
            }
        }
        // clear();
        if (rettype == NODE_TYPE) {
            tmplist = new ArrayList(1);
            tmplist.add(element);
            setOutput(tmplist);
        } else {
            setOutput(buf.toString());
        }
    }

    // public void visit(ExternalVariable arg) throws XQueryException {
    // Variable var = (Variable) arg.getObjectValue();
    // if (var.getExpression() != null)
    // var.getExpression().accept(this);
    // else
    // var.accept(this);
    // }

    public void visit(ITEExpression arg) throws XQueryException {
        returnType.push(BOOLEAN_TYPE);
        arg.getIfExpression().accept(this);
        returnType.pop();
        if (resverdict)
            arg.getThenExpression().accept(this);
        else
            arg.getElseExpression().accept(this);
    }

    public void visit(LibraryFunctionCall arg) throws XQueryException {
        FunctionDeclaration funDecl = arg.getFunctionDefinition();
        if (funDecl.getExpression() != null) {
            // TODO
        } else {
            Object[] params = null;
            if (arg.getArguments() != null) {
                params = new Object[arg.getArguments().size()];
                for (int i = 0; i < arg.getArguments().size(); i++) {
                    returnType.push(VALUE_TYPE);
                    arg.getArgument(i).accept(this);
                    params[i] = resvalue;
                    returnType.pop();
                }
            } else {
                params = new Object[0];
            }
            Object obj = dc.invoke(arg, params);
            Integer rettype = returnType.isEmpty() ? UNKNOWN_TYPE : (Integer) returnType.peek();
            if (rettype == NODE_TYPE) {
                ArrayList nodes = null;
                if (obj != null) {
                    nodes = new ArrayList(1);
                    Text text = docimpl.createTextNode(obj.toString());
                    nodes.add(text);
                }
                setOutput(nodes);
                return;
            } else {
                setOutput((Comparable) obj);
            }
        }
    }

    public void visit(ListOpArithExpression arg) throws XQueryException {
        XQueryExpression exp1 = arg.getExpression1();
        XQueryExpression exp2 = arg.getExpression2();
        int operator = arg.getOperator();

        returnType.push(UNKNOWN_TYPE);
        exp1.accept(this);
        ArrayList trees1 = resnodes;
        Number value1 = (resvalue instanceof Number) ? (Number) resvalue : null;
        exp2.accept(this);
        ArrayList trees2 = resnodes;
        Number value2 = (resvalue instanceof Number) ? (Number) resvalue : null;
        returnType.pop();

        if ((trees1 == null || trees1.isEmpty()) && (trees2 == null || trees2.isEmpty())) {
            Number res = processOperation(exp1, exp2, null, null, value1, value2, operator);
            SimpleType st = getSimpleType((Comparable) res, schemamanager.getDoubleType());
            TypedNode text = (TypedNode) docimpl.createTypedValue(res, st);
            text.setTypedValue(res);
            text.setType(st);
            ArrayList tmplist = new ArrayList(1);
            tmplist.add(text);
            setOutput(tmplist);
            return;
        }

        if ((trees1 == null || trees1.isEmpty()) && (trees2 != null) && (!trees2.isEmpty())) {
            ArrayList tmplist = new ArrayList(trees2.size());
            for (int j = 0; j < trees2.size(); j++) {
                TypedNode tree2 = (TypedNode) trees2.get(j);
                Number res = processOperation(exp1, exp2, null, tree2, value1, value2, operator);
                SimpleType st = getSimpleType((Comparable) res, schemamanager.getDoubleType());
                TypedNode text = (TypedNode) docimpl.createTypedValue(res, st);
                text.setTypedValue(res);
                text.setType(st);
                tmplist.add(text);
            }
            setOutput(tmplist);
            return;
        }

        if ((trees1 != null) && (!trees1.isEmpty()) && (trees2 == null) || (trees2.isEmpty())) {
            ArrayList tmplist = new ArrayList(trees1.size());
            for (int i = 0; i < trees1.size(); i++) {
                TypedNode tree1 = (TypedNode) trees1.get(i);
                Number res = processOperation(exp1, exp2, tree1, null, value1, value2, operator);
                SimpleType st = getSimpleType((Comparable) res, schemamanager.getDoubleType());
                TypedNode text = (TypedNode) docimpl.createTypedValue(res, st);
                text.setTypedValue(res);
                text.setType(st);
                tmplist.add(text);
            }
            setOutput(tmplist);
            return;
        }

        ArrayList tmplist = new ArrayList(trees1.size() * trees2.size());
        for (int i = 0; i < trees1.size(); i++) {
            TypedNode tree1 = (TypedNode) trees1.get(i);
            for (int j = 0; j < trees2.size(); j++) {
                TypedNode tree2 = (TypedNode) trees2.get(j);
                Number res = processOperation(exp1, exp2, tree1, tree2, value1, value2, operator);
                SimpleType st = getSimpleType((Comparable) res, schemamanager.getDoubleType());
                TypedNode text = (TypedNode) docimpl.createTypedValue(res, st);
                text.setTypedValue(res);
                text.setType(st);
                tmplist.add(text);
            }
        }
        setOutput(tmplist);
    }

    public void visit(ListOpCompExpression arg) throws XQueryException {
        XQueryExpression exp1 = arg.getExpression1();
        XQueryExpression exp2 = arg.getExpression2();
        int operator = arg.getOperator();

        returnType.push(EvaluationVisitor.NODE_TYPE);
        resnodes = null;
        resvalue = null;
        exp1.accept(this);
        ArrayList trees1 = resnodes;
        if (trees1 != null && trees1.isEmpty())
            trees1 = null;
        Comparable value1 = resvalue;
        resnodes = null;
        resvalue = null;
        exp2.accept(this);
        returnType.pop();
        ArrayList trees2 = resnodes;
        if (trees2 != null && trees2.isEmpty())
            trees2 = null;
        Comparable value2 = resvalue;

        if (value1 != null && value2 != null) {
            returnType.push(BOOLEAN_TYPE);
            processComparison(null, null, null, null, value1, value2, operator);
            returnType.pop();
            setOutput(resverdict);
            return;
        }

        if (trees1 == null && trees2 == null) {
            processComparison(exp1, exp2, null, null, value1, value2, operator);
            return;
        }

        if (trees1 == null && trees2 != null) {
            for (int j = 0; j < trees2.size(); j++) {
                TypedNode tree2 = (TypedNode) trees2.get(j);
                returnType.push(BOOLEAN_TYPE);
                processComparison(exp1, exp2, null, tree2, value1, value2, operator);
                returnType.pop();
                setOutput(resverdict);
                if (resverdict == true) {
                    setOutput(true);
                    return;
                }
            }
            setOutput(false);
            return;
        }

        if (trees1 != null && trees2 == null) {
            for (int i = 0; i < trees1.size(); i++) {
                TypedNode tree1 = (TypedNode) trees1.get(i);
                returnType.push(BOOLEAN_TYPE);
                processComparison(exp1, exp2, tree1, null, value1, value2, operator);
                returnType.pop();
                if (resverdict == true) {
                    setOutput(true);
                    return;
                }
            }
            setOutput(false);
            return;
        }

        for (int i = 0; i < trees1.size(); i++) {
            TypedNode tree1 = (TypedNode) trees1.get(i);
            for (int j = 0; j < trees2.size(); j++) {
                TypedNode tree2 = (TypedNode) trees2.get(j);
                returnType.push(BOOLEAN_TYPE);
                processComparison(exp1, exp2, tree1, tree2, value1, value2, operator);
                returnType.pop();
                if (resverdict == true) {
                    setOutput(true);
                    return;
                }
            }
            setOutput(false);
        }
    }

    public void visit(ListOpUNIONExpression arg) throws XQueryException {
        XQueryExpression exp1 = arg.getExpression1();
        XQueryExpression exp2 = arg.getExpression2();

        returnType.push(NODE_TYPE);
        resnodes = null;
        exp1.accept(this);
        ArrayList trees = resnodes;
        resnodes = null;
        exp2.accept(this);
        returnType.pop();
        if (resnodes != null) {
            if (trees == null)
                trees = resnodes;
            else
                trees.addAll(resnodes);
        }
        setOutput(trees);
    }

    public void visit(LocatedExpression arg) throws XQueryException {
        if (resultSet != null) {
            ArrayList steps = arg.getSteps();
            int nodeAccessor = arg.getNodeAccessor();
            boolean hasSpecialStep = false;
            if (nodeAccessor == Step.NONE) {
                int lastPosition = steps.size() - 1;
                Step lastStep = (Step) steps.get(lastPosition);
                if (lastStep.getStepKind() % 16 != 0) {
                    arg.setNodeAccessor(lastStep.getStepKind());
                    nodeAccessor = arg.getNodeAccessor();
                    steps.remove(lastPosition);
                }
            }
            clear();
            String str = null;
            try {
                if (generate) {
                    resultSet.generate(arg.getStringValue(), nodeAccessor);
                    return;
                } else
                    str = resultSet.fetch(arg.getStringValue(), nodeAccessor);
            } catch (XMLDBCException xe) {
                throw new XQueryException(xe.getMessage(), xe);
            }
            Integer rettype = returnType.isEmpty() ? UNKNOWN_TYPE : (Integer) returnType.peek();
            if (rettype == NODE_TYPE) {
                ArrayList nodes = null;
                if (str != null) {
                    nodes = new ArrayList(1);
                    Text text = docimpl.createTextNode(str);
                    nodes.add(text);
                }
                setOutput(nodes);
                return;
            } else {
                setOutput(str);
                return;
            }
        } else {
            clear();
            if (!(arg.getStepNum(0).getExpression() instanceof Variable)) {
                returnType.push(NODE_TYPE);
                arg.getStepNum(0).getExpression().accept(this);
                returnType.pop();
                ArrayList tmplist = resnodes;
                for (int j = 1; j < arg.getSteps().size(); j++) {
                    tmplist = DOMUtils.getStep(tmplist, arg.getStepNum(j));
                    if (tmplist == null || tmplist.isEmpty())
                        break;
                }
                if (tmplist != null)
                    setOutput(tmplist);
                return;
            }
            if (this.tuple == null && this.tuplelist == null) {
                return;
            }
            if (tuple != null) {
                if ((resnodes = tuple.getPath(arg.getStringValue())) != null) {
                    setOutput(resnodes);
                    return;
                }
            } else {
                for (int i = 0; i < tuplelist.size(); i++) {
                    if ((resnodes = ((Tuple) tuplelist.get(i)).getPath(arg.getStringValue())) != null) {
                        setOutput(resnodes);
                        return;
                    }
                }
            }
            resnodes = null;
            ArrayList steps = arg.getSteps();
            ArrayList tmplist = null;
            if (tuple != null) {
                StringBuffer buf = new StringBuffer(arg.toString(null, false, false, true));
                int i;
                for (i = steps.size() - 1; i > 0; i--) {
                    buf.delete(buf.length() - ((Step) steps.get(i)).getStringValue().length(), buf.length());
                    tmplist = tuple.getPath(buf.toString());
                    if (tmplist != null)
                        break;
                }
                if (tmplist == null)
                    return;
                for (int j = i; j < steps.size(); j++) {
                    tmplist = DOMUtils.getStep(tmplist, arg.getStepNum(j));
                    if (tmplist == null || tmplist.isEmpty())
                        break;
                }
            } else { // tuplelist not null
                for (int k = 0; k < tuplelist.size(); k++) {
                    StringBuffer buf = new StringBuffer(arg.getStringValue());
                    Tuple tuplek = (Tuple) tuplelist.get(k);
                    int i;
                    for (i = steps.size() - 1; i > 0; i--) {
                        buf.delete(buf.length() - ((Step) steps.get(i)).getStringValue().length(), buf.length());
                        tmplist = tuplek.getPath(buf.toString());
                        if (tmplist != null)
                            break;
                    }
                    if (tmplist == null)
                        continue;
                    for (int j = i; j < steps.size(); j++) {
                        tmplist = DOMUtils.getStep(tmplist, arg.getStepNum(j));
                        if (tmplist == null || tmplist.isEmpty())
                            break;
                    }
                    break;
                }
            }
            if (tmplist != null)
                setOutput(tmplist);
        }
    }

    public void visit(QName arg) throws XQueryException {
        String str = arg.getStringValue();
        Integer rettype = returnType.isEmpty() ? UNKNOWN_TYPE : (Integer) returnType.peek();
        if (rettype == NODE_TYPE) {
            ArrayList nodes = null;
            if (str != null) {
                nodes = new ArrayList(1);
                Text text = docimpl.createTextNode(str);
                nodes.add(text);
            }
            setOutput(nodes);
            return;
        } else {
            setOutput(str);
            return;
        }
    }

    private Comparable getMinus(Comparable obj) throws XQueryException {
        Comparable retObj = null;
        if (obj instanceof Number) {
            // BigDecimal, BigInteger, Byte, Double, Float, Integer, Long, Short
            if (obj instanceof Double) {
                retObj = new Double(-((Double) obj).doubleValue());
            } else if (obj instanceof Float) {
                retObj = new Float(-((Float) obj).floatValue());
            } else if (obj instanceof BigDecimal) {
                retObj = new BigDecimal(((BigDecimal) obj).unscaledValue().negate(), ((BigDecimal) obj).scale());
            } else if (obj instanceof Integer) {
                retObj = new Integer(-((Integer) obj).intValue());
            } else if (obj instanceof Long) {
                retObj = new Long(-((Long) obj).longValue());
            }
        } else if (obj instanceof String) {
            try {
                retObj = new Double(-Double.parseDouble((String) obj));
            } catch (NumberFormatException e) {
                // do nothing
            }
        }
        return retObj;
    }

    public void visit(UnOpMinusExpression arg) throws XQueryException {
        returnType.push(UNKNOWN_TYPE);
        arg.getExpression().accept(this);
        returnType.pop();
        Integer rettype = (Integer) returnType.peek();
        if (resvalue != null) {
            if (!arg.getMinus()) {
                if (rettype == NODE_TYPE) {
                    ArrayList nodes = new ArrayList(1);
                    TypedValue text = null;
                    if (resvalue instanceof Long)
                        text = docimpl.createTypedValue(resvalue, this.schemamanager.getIntegerType());
                    else if (resvalue instanceof BigDecimal)
                        text = docimpl.createTypedValue(resvalue, this.schemamanager.getDecimalType());
                    else if (resvalue instanceof Double)
                        text = docimpl.createTypedValue(resvalue, this.schemamanager.getDoubleType());
                    else
                        throw new XQueryException("UnOpMinusExpression could not cast resvalue");
                    nodes.add(text);
                    setOutput(nodes);
                    return;
                } else {
                    setOutput(resvalue);
                    return;
                }
            }
            Comparable comp = getMinus(resvalue);
            if (comp != null) {
                if (rettype == NODE_TYPE) {
                    ArrayList nodes = new ArrayList(1);
                    TypedValue text = null;
                    if (comp instanceof Long)
                        text = docimpl.createTypedValue(comp, this.schemamanager.getIntegerType());
                    else if (resvalue instanceof BigDecimal)
                        text = docimpl.createTypedValue(comp, this.schemamanager.getDecimalType());
                    else if (comp instanceof Double)
                        text = docimpl.createTypedValue(comp, this.schemamanager.getDoubleType());
                    else
                        throw new XQueryException("UnOpMinusExpression could not cast comp");
                    nodes.add(text);
                    setOutput(nodes);
                    return;
                } else {
                    setOutput(comp);
                    return;
                }
            } else
                throw new XQueryException("UnOpMinusExpression could not cast operator to Number");
        } else if (resnodes != null) {
            if (rettype == VALUE_TYPE && resnodes.size() != 1) {
                throw new XQueryException("UnOpMinusExpression could not get value from node list");
            }
            if (!arg.getMinus()) {
                if (rettype == NODE_TYPE) {
                    ArrayList tmplist = new ArrayList(resnodes.size());
                    for (int i = 0; i < resnodes.size(); i++) {
                        TypedNode tmpnode = (TypedNode) ((TypedNode) resnodes.get(0)).getFirstChild();
                        if (tmpnode == null)
                            tmpnode = (TypedNode) resnodes.get(0);
                        tmplist.add(docimpl.createTypedValue((Comparable) tmpnode.getTypedValue(), (SimpleType) tmpnode.getType()));
                    }
                    setOutput(tmplist);
                    return;
                } else { // (rettype == VALUE_TYPE) {
                    TypedNode tmpnode = (TypedNode) ((TypedNode) resnodes.get(0)).getFirstChild();
                    if (tmpnode == null)
                        tmpnode = (TypedNode) resnodes.get(0);
                    setOutput((Comparable) tmpnode.getTypedValue());
                    return;
                }
            }
            ArrayList tmplist = new ArrayList(resnodes.size());
            for (int i = 0; i < resnodes.size(); i++) {
                TypedNode tmpnode = (TypedNode) ((TypedNode) resnodes.get(0)).getFirstChild();
                if (tmpnode == null)
                    tmpnode = (TypedNode) resnodes.get(0);
                Comparable comp = getMinus((Comparable) tmpnode.getTypedValue());
                if (comp != null) {
                    TypedValue text = docimpl.createTypedValue(comp, (SimpleType) tmpnode.getType());
                    tmplist.add(text);
                }
            }
            if (rettype == NODE_TYPE) {
                setOutput(tmplist);
                return;
            } else {
                if (tmplist.size() == 1) {
                    setOutput((Comparable) ((TypedValue) tmplist.get(0)).getTypedValue());
                    return;
                }
            }
        }
        clear();
    }

    public void visit(Value arg) throws XQueryException {
        clear();
        String str = null;
        try {
            str = resultSet.fetch(arg.getStringValue(), 0);

        } catch (XMLDBCException xe) {
            throw new XQueryException(xe.getMessage(), xe);
        }
        Integer rettype = returnType.isEmpty() ? UNKNOWN_TYPE : (Integer) returnType.peek();
        ArrayList nodes = null;
        if (rettype == NODE_TYPE) {
            if (str != null) {
                nodes = new ArrayList(1);
                Text text = docimpl.createTextNode(str);
                nodes.add(text);
            }
            setOutput(nodes);
            return;
        } else {
            setOutput(str);
            return;
        }
    }

    // public void visit(ValueBoolean arg) throws XQueryException {
    public void visit(ValueDecimal arg) throws XQueryException {
        setOutput(arg.getDecimalValue());
    }

    public void visit(ValueDouble arg) throws XQueryException {
        setOutput(arg.getDoubleValue());
    }

    public void visit(ValueFloat arg) throws XQueryException {
        setOutput(arg.getFloatValue());
    }

    public void visit(ValueInteger arg) throws XQueryException {
        setOutput(new Long(arg.getIntValue().longValue()));
    }

    public void visit(ValueString arg) throws XQueryException {
        setOutput(arg.getStringValue());
    }

    public void visit(ValueText arg) throws XQueryException {
        setOutput(arg.getTextValue());
    }

    public void visit(Variable arg) throws XQueryException {
        if (arg.isExternal() && arg.getExternalValue() != null) {
            clear();
            Integer rettype = returnType.isEmpty() ? UNKNOWN_TYPE : (Integer) returnType.peek();
            if (rettype == NODE_TYPE) {
                ArrayList nodes = new ArrayList(1);
                TypedValue text = docimpl.createTypedValue(arg.getExternalValue().toString(), this.schemamanager.getStringType());
                nodes.add(text);
                setOutput(nodes);
                return;
            } else {
                setOutput(arg.getExternalValue().toString());
                return;
            }
        } else if (resultSet != null) {
            clear();
            String str = null;
            try {
                if (generate) {
                    resultSet.generate(arg.getStringValue(), 0);
                    return;
                } else
                    str = resultSet.fetch(arg.getStringValue(), 0);

            } catch (XMLDBCException xe) {
                throw new XQueryException(xe.getMessage(), xe);
            }
            Integer rettype = returnType.isEmpty() ? UNKNOWN_TYPE : (Integer) returnType.peek();
            if (rettype == NODE_TYPE) {
                ArrayList nodes = null;
                if (str != null) {
                    nodes = new ArrayList(1);
                    Text text = docimpl.createTextNode(str);
                    nodes.add(text);
                }
                setOutput(nodes);
                return;
            } else {
                setOutput(str);
                return;
            }
        } else if (this.tuple != null) {
            setOutput(tuple.getPath(arg.getStringValue()));
            return;
        } else if (tuplelist != null) {
            for (int i = 0; i < tuplelist.size(); i++) {
                if ((resnodes = ((Tuple) tuplelist.get(i)).getPath(arg.getStringValue())) != null) {
                    setOutput(resnodes);
                    return;
                }
            }
        }
        this.clear();
    }

    // public void visit (XQueryBinaryOperatorExpression arg) throws
    // XQueryException
    // public void visit(XQueryBooleanOperatorExpression arg) throws
    // XQueryException
    public void visit(XQueryExpression arg) throws XQueryException {
        throw new XQueryException("Class " + arg.getClass().getName() + " not supported in EvaluationVisitor");
    }

    public void visit(XQueryExpressionSequence arg) throws XQueryException {
        if (arg.getSubExpressions().size() == 1)
            ((XQueryExpression) arg.getSubExpressions().get(0)).accept(this);
        else {
            ArrayList expressions = arg.getSubExpressions();
            Integer rettype = returnType.isEmpty() ? UNKNOWN_TYPE : (Integer) returnType.peek();
            if (rettype == VALUE_TYPE) {
                StringBuffer buf = new StringBuffer();
                returnType.push(VALUE_TYPE);
                for (int i = 0; i < expressions.size(); i++) {
                    XQueryExpression expri = (XQueryExpression) expressions.get(i);
                    expri.accept(this);
                    if (resvalue != null) {
                        if (buf.length() != 0)
                            buf.append(' ');
                        buf.append(resvalue.toString());
                    }
                }
                setOutput(buf.toString());
                returnType.pop();
            } else {
                ArrayList tmpnodes = new ArrayList();
                returnType.push(NODE_TYPE);
                for (int i = 0; i < expressions.size(); i++) {
                    XQueryExpression expri = (XQueryExpression) expressions.get(i);
                    expri.accept(this);
                    if (resnodes != null)
                        tmpnodes.addAll(resnodes);
                }
                setOutput(tmpnodes);
                returnType.pop();
            }
        }
    }

    public void visit(ParenthesizedExpression arg) throws XQueryException {
        XQueryExpression xqexpr = arg.getExpression();
        if (xqexpr == null) {
            clear();
            return;
        }
        xqexpr.accept(this);
    }

    public void visit(EnclosedExpression arg) throws XQueryException {
        XQueryExpression xqexpr = arg.getExpression();
        if (xqexpr == null) {
            clear();
            return;
        }
        xqexpr.accept(this);
    }

    // -----------------------------------------------------------
    // BOOLEAN FUNCTIONS
    // -----------------------------------------------------------
    public void visit(FunctionFALSE arg) throws XQueryException {
        setOutput(false);
    }

    public void visit(FunctionTRUE arg) throws XQueryException {
        setOutput(true);
    }

    // -----------------------------------------------------------
    // AGGREGATE FUNCTIONS
    // -----------------------------------------------------------
    private boolean isLongSum = false;

    private double doubleSum = 0;

    private long longSum = 0;

    private long count = 0;

    private Comparable comp = null;

    TypedNode result = null;

    public void visit(FunctionAVG arg) throws XQueryException {
        if (incrementalAggregate && tuple != null) {
            aggKind = AVG;
            int index = tuple.getParent().getPathIndex(arg.getArgument(0).getStringValue());
            if (index == -1)
                index = 0;
            Number valuei = null;
            ArrayList list = (ArrayList) tuple.getNodesAtIndex(index);
            if (list == null)
                return;
            for (int i = 0; i < list.size(); i++) {
                TypedNode treei = (TypedNode) list.get(i);
                try {
                    if (treei.getTypedValue() != null)
                        valuei = (Number) treei.getTypedValue();
                    else
                        valuei = (Number) DOMUtils.getTextLeaf(treei);
                } catch (ClassCastException cce) {
                    // do nothing
                }
                if (valuei == null)
                    return;
                if (isLongSum && (valuei instanceof Integer || valuei instanceof Long)) {
                    longSum += valuei.longValue();
                } else {
                    if (isLongSum) {
                        doubleSum = longSum;
                        isLongSum = false;
                    }
                    doubleSum += valuei.doubleValue();
                }
                count++;
            }
            // do not build return value;
            return;
        } else {
            returnType.push(NODE_TYPE);
            arg.getArgument(0).accept(this);
            returnType.pop();
            if (resnodes == null || resnodes.isEmpty())
                return;

            for (int i = 0; i < resnodes.size(); i++) {
                Number valuei = null;
                TypedNode treei = (TypedNode) resnodes.get(i);
                try {
                    if (treei.getTypedValue() != null)
                        valuei = (Number) treei.getTypedValue();
                    else
                        valuei = (Number) DOMUtils.getTextLeaf(treei);
                } catch (ClassCastException cce) {
                    // do nothing
                }
                if (valuei == null)
                    return;
                if (isLongSum && (valuei instanceof Integer || valuei instanceof Long)) {
                    longSum += valuei.longValue();
                } else {
                    if (isLongSum) {
                        doubleSum = longSum;
                        isLongSum = false;
                    }
                    doubleSum += valuei.doubleValue();
                }
            }
            count = resnodes.size();
        }
        if (!returnType.isEmpty()) {
            Integer retType = (Integer) this.returnType.peek();
            if (retType == NODE_TYPE) {
                ArrayList nodes = new ArrayList(1);
                TypedValue text = null;
                if (isLongSum)
                    text = docimpl.createTypedValue(new BigDecimal(longSum / count), this.schemamanager.getDecimalType());
                else
                    text = docimpl.createTypedValue(new Double(doubleSum / count), this.schemamanager.getDoubleType());
                nodes.add(text);
                setOutput(nodes);
                return;
            } else {
                if (isLongSum)
                    setOutput(new BigDecimal(longSum / count));
                else
                    setOutput(new Double(doubleSum / count));
                return;
            }
        }
        // clear();

    }

    public void visit(FunctionSUM arg) throws XQueryException {
        // treat special case first
        if (incrementalAggregate && tuple != null) {
            aggKind = SUM;
            int index = tuple.getParent().getPathIndex(arg.getArgument(0).getStringValue());
            if (index == -1)
                index = 0;
            Number valuei = null;
            ArrayList list = (ArrayList) tuple.getNodesAtIndex(index);
            if (list == null)
                return;
            for (int i = 0; i < list.size(); i++) {
                TypedNode treei = (TypedNode) list.get(i);
                try {
                    if (treei.getTypedValue() != null)
                        valuei = (Number) treei.getTypedValue();
                    else
                        valuei = (Number) DOMUtils.getTextLeaf(treei);
                } catch (ClassCastException cce) {
                    // do nothing
                }
                if (valuei == null)
                    return;
                if (isLongSum && (valuei instanceof Integer || valuei instanceof Long)) {
                    longSum += valuei.longValue();
                } else {
                    if (isLongSum) {
                        doubleSum = longSum;
                        isLongSum = false;
                    }
                    doubleSum += valuei.doubleValue();
                }
            }
            // do not build return value
            return;
        } else {
            returnType.push(NODE_TYPE);
            arg.getArgument(0).accept(this);
            returnType.pop();
            if (resnodes == null || resnodes.isEmpty())
                return;

            for (int i = 0; i < resnodes.size(); i++) {
                Number valuei = null;
                TypedNode treei = (TypedNode) resnodes.get(i);
                try {
                    if (treei.getTypedValue() != null)
                        valuei = (Number) treei.getTypedValue();
                    else
                        valuei = (Number) DOMUtils.getTextLeaf(treei);
                } catch (ClassCastException cce) {
                    // do nothing
                }
                if (valuei == null)
                    return;
                if (isLongSum && (valuei instanceof Integer || valuei instanceof Long)) {
                    longSum += valuei.longValue();
                } else {
                    if (isLongSum) {
                        doubleSum = longSum;
                        isLongSum = false;
                    }
                    doubleSum += valuei.doubleValue();
                }
            }
        }
        if (!returnType.isEmpty()) {
            Integer retType = (Integer) this.returnType.peek();
            if (retType == NODE_TYPE) {
                ArrayList nodes = new ArrayList(1);
                TypedValue text = null;
                if (isLongSum)
                    text = docimpl.createTypedValue(new Long(longSum), this.schemamanager.getIntegerType());
                else
                    text = docimpl.createTypedValue(new Double(doubleSum), this.schemamanager.getDoubleType());
                nodes.add(text);
                setOutput(nodes);
                return;
            } else {
                if (isLongSum)
                    setOutput(new Long(longSum));
                else
                    setOutput(new Double(doubleSum));
                return;
            }
        }
        // clear();

    }

    public void visit(FunctionMAX arg) throws XQueryException {
        if (incrementalAggregate && tuple != null) {
            aggKind = MAX;
            int index = tuple.getParent().getPathIndex(arg.getArgument(0).getStringValue());
            if (index == -1)
                index = 0;
            Comparable valuei = null;
            ArrayList list = (ArrayList) tuple.getNodesAtIndex(index);
            TypedNode treei = null;

            if (list == null)
                return;
            for (int i = 0; i < list.size(); i++) {
                treei = (TypedNode) list.get(i);
                try {
                    if (treei.getTypedValue() != null)
                        valuei = (Comparable) treei.getTypedValue();
                    else
                        valuei = (Comparable) DOMUtils.getTextLeaf(treei);
                } catch (ClassCastException cce) {
                    // do nothing
                }
                if (valuei == null)
                    return;
                if (comp == null) {
                    comp = valuei;
                    result = treei;
                } else if (DOMUtils.compareObjects(valuei, comp) > 0) {
                    comp = valuei;
                    result = treei;
                }
            }
            // do not build result value
            return;
        } else {
            returnType.push(NODE_TYPE);
            arg.getArgument(0).accept(this);
            returnType.pop();
            if (resnodes == null || resnodes.isEmpty())
                return;

            for (int i = 0; i < resnodes.size(); i++) {
                Comparable valuei = null;
                TypedNode treei = (TypedNode) resnodes.get(i);
                try {
                    if (treei.getTypedValue() != null)
                        valuei = (Comparable) treei.getTypedValue();
                    else
                        valuei = (Comparable) DOMUtils.getTextLeaf(treei);
                } catch (ClassCastException cce) {
                    // do nothing
                }
                if (valuei == null)
                    return;
                if (comp == null) {
                    comp = valuei;
                    result = treei;
                } else if (DOMUtils.compareObjects(valuei, comp) > 0) {
                    comp = valuei;
                    result = treei;
                }
            }
        }
        if (!returnType.isEmpty()) {
            Integer retType = (Integer) this.returnType.peek();
            if (retType == NODE_TYPE) {
                ArrayList nodes = new ArrayList(1);
                TypedValue text = docimpl.createTypedValue(comp, (SimpleType) result.getType());
                nodes.add(text);
                setOutput(nodes);
                return;
            } else {
                setOutput(comp);
                return;
            }
        }
        // clear();

    }

    public void visit(FunctionMIN arg) throws XQueryException {
        if (incrementalAggregate && tuple != null) {
            aggKind = MIN;
            int index = tuple.getParent().getPathIndex(arg.getArgument(0).getStringValue());
            if (index == -1)
                index = 0;
            Comparable valuei = null;
            ArrayList list = (ArrayList) tuple.getNodesAtIndex(index);
            TypedNode treei = null;
            if (list == null)
                return;
            for (int i = 0; i < list.size(); i++) {
                treei = (TypedNode) list.get(i);
                try {
                    if (treei.getTypedValue() != null)
                        valuei = (Comparable) treei.getTypedValue();
                    else
                        valuei = (Comparable) DOMUtils.getTextLeaf(treei);
                } catch (ClassCastException cce) {
                    // do nothing
                }
                if (valuei == null)
                    return;
                if (comp == null) {
                    comp = valuei;
                    result = treei;
                } else if (DOMUtils.compareObjects(valuei, comp) < 0) {
                    comp = valuei;
                    result = treei;
                }
            }
            // do not build result value
            return;
        } else {
            returnType.push(NODE_TYPE);
            arg.getArgument(0).accept(this);
            returnType.pop();
            if (resnodes == null || resnodes.isEmpty())
                return;

            for (int i = 0; i < resnodes.size(); i++) {
                Comparable valuei = null;
                TypedNode treei = (TypedNode) resnodes.get(i);
                try {
                    if (treei.getTypedValue() != null)
                        valuei = (Comparable) treei.getTypedValue();
                    else
                        valuei = (Comparable) DOMUtils.getTextLeaf(treei);
                } catch (ClassCastException cce) {
                    // do nothing
                }
                if (valuei == null)
                    return;
                if (comp == null) {
                    comp = valuei;
                    result = treei;
                } else if (DOMUtils.compareObjects(valuei, comp) < 0) {
                    comp = valuei;
                    result = treei;
                }
            }
        }
        if (!returnType.isEmpty()) {
            Integer retType = (Integer) this.returnType.peek();
            if (retType == NODE_TYPE) {
                ArrayList nodes = new ArrayList(1);
                TypedValue text = docimpl.createTypedValue(comp, (SimpleType) result.getType());
                nodes.add(text);
                setOutput(nodes);
                return;
            } else {
                setOutput(comp);
                return;
            }
        }
        // clear();

    }

    public void visit(FunctionCOUNT arg) throws XQueryException {
        if (incrementalAggregate && tuple != null) {
            aggKind = COUNT;
            count++;
            return;
        } else {
            returnType.push(NODE_TYPE);
            arg.getArgument(0).accept(this);
            returnType.pop();
            if (resnodes != null)
                count = resnodes.size();
        }
        if (!returnType.isEmpty()) {
            Integer retType = (Integer) this.returnType.peek();
            if (retType == NODE_TYPE) {
                ArrayList nodes = new ArrayList(1);
                TypedValue text = docimpl.createTypedValue(new Long(count), this.schemamanager.getIntegerType());
                nodes.add(text);
                setOutput(nodes);
                return;
            } else {
                setOutput(new Long(count));
                return;
            }
        }
        // clear();
    }

    // -----------------------------------------------------------
    // DATE/TIME FUNCTIONS
    // -----------------------------------------------------------
    // private int[] parseDate(String date) {
    // if (date == null)
    // return null;
    // int[] retval = new int[3];
    // retval[0] = Integer.parseInt(date.substring(0, 4));
    // retval[1] = Integer.parseInt(date.substring(5, 7));
    // retval[2] = Integer.parseInt(date.substring(8, 10));
    // return retval;
    // }

    // Calendar cal = Calendar.getInstance();

    public void visit(FunctionCURRENT_DATETIME arg) throws XQueryException {
        setOutput(dc.getCurrentDateTime());
    }

    public void visit(FunctionDAYTIMEDURATION arg) throws XQueryException {
        returnType.push(VALUE_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        if (resvalue instanceof Duration) { // do nothing
        } else if (resvalue instanceof String) {
            try {
                resvalue = (Comparable) this.schemamanager.getDurationType().convert((String) resvalue, true, null);
            } catch (SchemaException se) {
                throw new XQueryException("Argument of function dayTimeDuration is not a valid duration");
            }
        } else
            throw new XQueryException("Argument of function dayTimeDuration cannot be converted");
        setOutput(resvalue);
    }

    public void visit(FunctionDATE arg) throws XQueryException {
        returnType.push(VALUE_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        if (resvalue instanceof DateTime) { // do nothing
        } else if (resvalue instanceof String) {
            try {
                resvalue = (Comparable) this.schemamanager.getDateType().convert((String) resvalue, true, null);
            } catch (SchemaException se) {
                throw new XQueryException("Argument of function date is not a valid date");
            }
        } else
            throw new XQueryException("Argument of function date cannot be converted");
        // ((DateTime) resvalue).setHour(DatatypeConstants.FIELD_UNDEFINED);
        // ((DateTime) resvalue).setMinute(DatatypeConstants.FIELD_UNDEFINED);
        // ((DateTime) resvalue).setSecond(DatatypeConstants.FIELD_UNDEFINED);
        // ((DateTime) resvalue).setMillisecond(DatatypeConstants.FIELD_UNDEFINED);
        setOutput(resvalue);
    }

    public void visit(FunctionDATETIME arg) throws XQueryException {
        returnType.push(VALUE_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        if (resvalue instanceof DateTime) { // do nothing
        } else if (resvalue instanceof String) {
            try {
                resvalue = (Comparable) this.schemamanager.getDateTimeType().convert((String) resvalue, true, null);
            } catch (SchemaException se) {
                throw new XQueryException("Argument of function datetime is not a valid date");
            }
        } else
            throw new XQueryException("Argument of function datetime cannot be converted");
        setOutput(resvalue);
    }

    public void visit(FunctionTIME arg) throws XQueryException {
        returnType.push(VALUE_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        if (resvalue instanceof DateTime) { // do nothing
        } else if (resvalue instanceof String) {
            try {
                resvalue = (Comparable) this.schemamanager.getTimeType().convert((String) resvalue, true, null);
            } catch (SchemaException se) {
                throw new XQueryException("Argument of function time is not a valid date");
            }
        } else
            throw new XQueryException("Argument of function time cannot be converted");
        // ((DateTime) resvalue).setYear(DatatypeConstants.FIELD_UNDEFINED);
        // ((DateTime) resvalue).setMonth(DatatypeConstants.FIELD_UNDEFINED);
        // ((DateTime) resvalue).setDay(DatatypeConstants.FIELD_UNDEFINED);
        setOutput(resvalue);
    }

    public void visit(FunctionYEAR_FROM_DATE arg) throws XQueryException {
        returnType.push(VALUE_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        if (!(resvalue instanceof DateTime))
            throw new XQueryException("Argument of function get-year-from-date is not a date");
        setOutput(new Integer(((DateTime) resvalue).getYear()), schemamanager.getIntegerType());
    }

    public void visit(FunctionMONTH_FROM_DATE arg) throws XQueryException {
        returnType.push(VALUE_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        if (!(resvalue instanceof DateTime))
            throw new XQueryException("Argument of function get-year-from-date is not a date");
        setOutput(new Integer(((DateTime) resvalue).getMonth()), schemamanager.getIntegerType());
    }

    // private static Duration maxTimeZoneValue = new Duration("PT14H");
    //
    // private static Duration minTimeZoneValue = new Duration("-PT14H");

    private void adjustToTimezone(PrimitiveFunctionCall arg, String funcName, SimpleType simpleType) throws XQueryException {
        // ArrayList args = arg.getArguments();
        // returnType.push(VALUE_TYPE);
        // arg.getArgument(0).accept(this);
        // returnType.pop();
        // if (resvalue != null) {
        // if (resvalue instanceof DateTime) { // do nothing
        // } else if (resvalue instanceof String) {
        // try {
        // resvalue = (DateTime) this.schemamanager.getDateTimeType().convert((String) resvalue, true, null);
        // } catch (SchemaException se) {
        // throw new XQueryException("First argument of function " + funcName + " is not valid.");
        // }
        // } else
        // throw new XQueryException("First argument of function " + funcName + " cannot be converted.");
        // }
        // DateTime dateTime = (DateTime) resvalue;
        // resvalue = null;
        // Duration timezone = dc.getCurrentTimeZone();
        // if (args.size() > 1) {
        // returnType.push(VALUE_TYPE);
        // arg.getArgument(1).accept(this);
        // returnType.pop();
        // if (resvalue != null) {
        // if (resvalue instanceof Duration) {
        // timezone = (Duration) resvalue;
        // } else if (resvalue instanceof String) {
        // try {
        // timezone = new Duration((String) resvalue);
        // } catch (IllegalArgumentException iae) {
        // throw new XQueryException("Second argument of function " + funcName + " cannot be converted");
        // }
        // } else
        // throw new XQueryException("Second argument of function " + funcName + " cannot be converted");
        // } else
        // timezone = null;
        // }
        // if (dateTime != null) {
        // if (timezone != null) {
        // // check that timezone is betwween -14H and +14H
        // if (timezone.compare(maxTimeZoneValue) == DatatypeConstants.GREATER || timezone.compare(minTimeZoneValue) == DatatypeConstants.LESSER)
        // throw new XQueryException("Evaluation of function " + funcName + ", second argument timezone has incorrect value.");
        // if (timezone.isSet(DatatypeConstants.SECONDS)) {
        // Number num = timezone.getField(DatatypeConstants.SECONDS);
        // if (num.floatValue() > 0)
        // throw new XQueryException("Evaluation of function " + funcName + ", second argument timezone has incorrect value.");
        // }
        // // TimeZone timeZone = cal.getTimeZone();
        // int offset = (timezone.getHours() * 60 + timezone.getMinutes()) * timezone.getSign();
        // long dateTimeOffset = dateTime.getTimeZone();
        // if (dateTimeOffset != DatatypeConstants.FIELD_UNDEFINED)
        // dateTime.add(new Duration((offset - dateTimeOffset) * 60000));
        // dateTime.setTimezone(offset);
        // } else {
        // dateTime.setTimezone(DatatypeConstants.FIELD_UNDEFINED);
        // }
        // }
        // setOutput(dateTime, simpleType);
    }

    public void visit(FunctionADJUST_DATE_TO_TIMEZONE arg) throws XQueryException {
        adjustToTimezone(arg, "adjust-date-to-timezone", this.schemamanager.getDateType());
    }

    public void visit(FunctionADJUST_DATETIME_TO_TIMEZONE arg) throws XQueryException {
        adjustToTimezone(arg, "adjust-dateTime-to-timezone", this.schemamanager.getDateTimeType());
    }

    public void visit(FunctionADJUST_TIME_TO_TIMEZONE arg) throws XQueryException {
        adjustToTimezone(arg, "adjust-time-to-timezone", this.schemamanager.getTimeType());
    }

    private void getTimezone(PrimitiveFunctionCall arg, String funcName) throws XQueryException {
        returnType.push(VALUE_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        if (resvalue instanceof DateTime) { // do nothing
        } else if (resvalue instanceof String) {
            try {
                resvalue = (DateTime) this.schemamanager.getDateTimeType().convert((String) resvalue, true, null);
            } catch (SchemaException se) {
                throw new XQueryException("First argument of function " + funcName + " is not a valid.");
            }
        } else
            throw new XQueryException("First argument of function " + funcName + " cannot be converted.");
        setOutput(new Duration(1, 0, 0, 0, 0, 0, ((DateTime) resvalue).getTimeZone()), this.schemamanager.getDurationType());
    }

    public void visit(FunctionTIMEZONE_FROM_DATETIME arg) throws XQueryException {
        getTimezone(arg, "timezone-from-dateTime");
    }

    public void visit(FunctionTIMEZONE_FROM_DATE arg) throws XQueryException {
        getTimezone(arg, "timezone-from-date");
    }

    public void visit(FunctionTIMEZONE_FROM_TIME arg) throws XQueryException {
        getTimezone(arg, "timezone-from-tTime");
    }

    public void visit(FunctionIMPLICIT_TIMEZONE arg) throws XQueryException {
        setOutput(dc.getCurrentTimeZone());
    }

    // -----------------------------------------------------------
    // STRING FUNCTIONS
    // -----------------------------------------------------------

    private Comparable c0 = null;

    private Comparable c1 = null;

    private void getComparables(XQueryExpression arg0, XQueryExpression arg1) throws XQueryException {
        resnodes = null;
        c0 = null;
        c1 = null;
        if (arg0 instanceof Value)
            c0 = (((Value) arg0).getValue());
        if (arg1 instanceof Value)
            c1 = (((Value) arg1).getValue());
        if (c0 == null) {
            returnType.push(VALUE_TYPE);
            arg0.accept(this);
            returnType.pop();
            c0 = this.resvalue;
            if (c0 == null && resnodes != null && resnodes.size() == 1) {
                TypedNode node = (TypedNode) resnodes.get(0);
                if (node instanceof TypedValueImpl) {
                    c0 = (Comparable) node.getTypedValue();
                } else if (node instanceof TypedElementImpl) {
                    StringBuffer sb = new StringBuffer();
                    DOMUtils.getStringValue(node, sb);
                    c0 = sb.toString();
                }
            }
        }
        if (c1 == null) {
            returnType.push(VALUE_TYPE);
            arg1.accept(this);
            returnType.pop();
            c1 = this.resvalue;
            if (c1 == null && resnodes != null && resnodes.size() == 1) {
                TypedNode node = (TypedNode) resnodes.get(0);
                if (node instanceof TypedValueImpl) {
                    c1 = (Comparable) node.getTypedValue();
                } else if (node instanceof TypedElementImpl) {
                    StringBuffer sb = new StringBuffer();
                    DOMUtils.getStringValue(node, sb);
                    c1 = sb.toString();
                }
            }
        }

    }

    public void visit(FunctionLOWER_CASE arg) throws XQueryException {
        returnType.push(UNKNOWN_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        if (resvalue != null) { // do nothing
        } else if (resnodes != null && resnodes.size() == 1) {
            TypedNode tmpnode = (TypedNode) ((TypedNode) resnodes.get(0)).getFirstChild();
            if (tmpnode == null)
                tmpnode = (TypedNode) resnodes.get(0);
            resvalue = (Comparable) tmpnode.getTypedValue();
        }
        if (resvalue instanceof String) {
            resvalue = ((String) resvalue).toLowerCase();
        } else {
            throw new XQueryException("Argument of function lower-case is not a string");
        }
        Integer rettype = (Integer) returnType.peek();
        if (rettype == VALUE_TYPE) {
            setOutput(resvalue);
            return;
        } else if (rettype == NODE_TYPE) {
            ArrayList nodes = new ArrayList(1);
            nodes.add(docimpl.createTypedValue(resvalue, this.getSimpleType(resvalue, schemamanager.getStringType())));
            setOutput(nodes);
            return;
        }
    }

    public void visit(FunctionUPPER_CASE arg) throws XQueryException {
        returnType.push(UNKNOWN_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        if (resvalue != null) { // do nothing
        } else if (resnodes != null && resnodes.size() == 1) {
            TypedNode tmpnode = (TypedNode) ((TypedNode) resnodes.get(0)).getFirstChild();
            if (tmpnode == null)
                tmpnode = (TypedNode) resnodes.get(0);
            resvalue = (Comparable) tmpnode.getTypedValue();
        }
        if (resvalue instanceof String) {
            resvalue = ((String) resvalue).toUpperCase();
        } else {
            throw new XQueryException("Argument of function upper-case is not a string");
        }
        Integer rettype = (Integer) returnType.peek();
        if (rettype == VALUE_TYPE) {
            setOutput(resvalue);
            return;
        } else if (rettype == NODE_TYPE) {
            ArrayList nodes = new ArrayList(1);
            nodes.add(docimpl.createTypedValue(resvalue, this.getSimpleType(resvalue, schemamanager.getStringType())));
            setOutput(nodes);
            return;
        }
    }

    public void visit(FunctionCONCAT arg) throws XQueryException {
        StringBuffer buf = new StringBuffer();
        for (int i = 0; i < arg.getArguments().size(); i++) {
            returnType.push(UNKNOWN_TYPE);
            arg.getArgument(i).accept(this);
            returnType.pop();
            if (resvalue != null) { // do nothing
                if (resvalue instanceof String)
                    buf.append(resvalue);
                else
                    buf.append(resvalue.toString());
            } else if (resnodes == null || resnodes.isEmpty()) { // do
                // nothing
            } else if (resnodes != null && resnodes.size() == 1) {
                StringBuffer sb = new StringBuffer();
                DOMUtils.getStringValue((TypedNode) resnodes.get(0), sb);
                buf.append(sb.toString());
            } else if (resnodes != null && resnodes.size() > 1) {
                throw new XQueryException("Evaluation : incorrect argument for function concat");
            }
        }
        Integer rettype = (Integer) returnType.peek();
        if (rettype == NODE_TYPE) {
            ArrayList nodes = new ArrayList(1);
            nodes.add(docimpl.createTypedValue(buf.toString(), this.schemamanager.getStringType()));
            setOutput(nodes);
            return;
        } else {
            setOutput(buf.toString());
            return;
        }
        // clear();
    }

    public void visit(FunctionCONTAINS arg) throws XQueryException {
        getComparables(arg.getArgument(0), arg.getArgument(1));
        if (c0 != null && c1 != null) {
            setOutput(((String) c0).indexOf((String) c1) != -1);
            return;
        }
        clear();
    }

    public void visit(FunctionENDS_WITH arg) throws XQueryException {
        getComparables(arg.getArgument(0), arg.getArgument(1));
        if (c0 != null && c1 != null) {
            setOutput(((String) c0).endsWith((String) c1));
            return;
        }
        clear();
    }

    public void visit(FunctionSTARTS_WITH arg) throws XQueryException {
        getComparables(arg.getArgument(0), arg.getArgument(1));
        if (c0 != null && c1 != null) {
            setOutput(((String) c0).startsWith((String) c1));
            return;
        }
        clear();
    }

    public void visit(FunctionLOCAL_NAME arg) throws XQueryException {
        returnType.push(NODE_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        Integer rettype = (Integer) returnType.peek();
        // if (rettype == VALUE_TYPE) {
        if (this.resnodes != null && resnodes.size() == 1) {
            this.setOutput(((TypedNode) resnodes.get(0)).getLocalName());
            return;
        }
        // }
        clear();
    }

    public void visit(FunctionNAMESPACE_URI arg) throws XQueryException {
        returnType.push(NODE_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        Integer rettype = (Integer) returnType.peek();
        // if (rettype == VALUE_TYPE) {
        if (this.resnodes != null && resnodes.size() == 1) {
            this.setOutput(((TypedNode) resnodes.get(0)).getNamespaceURI());
            return;
        }
        // }
        clear();
    }

    /*
     * 2.3 fn:string fn:string() as xs:string fn:string($srcval as item?) as xs:string Returns the value of $srcval represented as a xs:string. If no argument is supplied, $srcval defaults to the context item (.).
     * 
     * If $srcval is the empty sequence, the zero-length string is returned.
     * 
     * If $srcval is a node, the function returns the string-value of the node, as obtained using the string-value accessor defined in the [XQuery 1.0 and XPath 2.0 Data Model].
     * 
     * If $srcval is an atomic value, then the function returns the same string as is returned by the expression cast as xs:string ($srcval), except in the cases listed below:
     * 
     * If the type of $srcval is xs:anyURI, the URI is converted to a string without any escaping of special characters.
     * 
     * Note:
     * 
     * The reason for the special rule for xs:anyURI is that, although XML Schema strongly discourages the use of spaces within URI values, the escaping of spaces can cause problems with legacy applications (for example, this applies to spaces within fragment identifiers in many HTML browsers), and should therefore be under user control.
     * 
     * Note:
     * 
     * The string representation of xs:double values is not backwards-compatible with the representation of number values in [XPath 1.0]. Ordinary xs:double values are now represented using scientific notation; the representations of positive and negative infinity are now 'INF' and '-INF' rather than 'Infinity' and '-Infinity'. (It should be observed that '+INF' is not supported as a lexical form of infinity in [XML Schema Part 2: Datatypes] and is thus not supported by this specification; if that lexical form is added in a future version of [XML Schema Part 2: Datatypes], then it will be supported by a future version of this specification that aligns with that future version of [XML Schema Part 2: Datatypes].) However, most expressions that would have produced a number in [XPath 1.0] will produce a decimal (or integer) in [XPath 2.0], so unless there is a loss of precision caused by numeric approximation, the result of the expression will in most simple cases be the same after
     * conversion to a string.
     */
    public void visit(FunctionSTRING arg) throws XQueryException {
        returnType.push(UNKNOWN_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        if (resvalue != null) { // do nothing
            if (!(resvalue instanceof String))
                resvalue = resvalue.toString();
        } else if (resnodes == null || resnodes.isEmpty()) {
            resvalue = new String("");
        } else if (resnodes != null && resnodes.size() == 1) {
            StringBuffer sb = new StringBuffer();
            DOMUtils.getStringValue((TypedNode) resnodes.get(0), sb);
            resvalue = sb.toString();
        } else if (resnodes != null && resnodes.size() > 1) {
            throw new XQueryException("Evaluation : incorrect argument for function string");
        }
        Integer rettype = (Integer) returnType.peek();
        if (rettype == NODE_TYPE) {
            ArrayList nodes = new ArrayList(1);
            TypedValue text = docimpl.createTypedValue(resvalue, this.schemamanager.getStringType());
            nodes.add(text);
            setOutput(nodes);
            return;
        } else {
            setOutput(resvalue);
            return;
        }
        // clear();
    }

    /*
     * 7.4.7 fn:string-length fn:string-length() as xs:integer? fn:string-length($srcval as xs:string?) as xs:integer? Returns an xs:integer equal to the length in characters of the value of $srcval. If the value of $srcval is the empty sequence, the empty sequence is returned. If no argument is supplied, $srcval defaults to the string value (calculated using fn:string()) of the context item (.).
     * 
     * Note:
     * 
     * The value returned is the number of XML characters in the string (or equivalently, the number of Unicode code points). Some implementations may represent a code point above xFFFF using two 16-bit values known as a surrogate pair. A surrogate pair counts as one character, not two.
     */
    public void visit(FunctionSTRING_LENGTH arg) throws XQueryException {
        returnType.push(UNKNOWN_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        if (resvalue != null) { // do nothing
            if (resvalue instanceof String)
                resvalue = new Long(((String) resvalue).length());
            else
                resvalue = new Long(resvalue.toString().length());
        } else if (resnodes == null || resnodes.isEmpty()) {
            resvalue = new Long(0);
        } else if (resnodes != null && resnodes.size() == 1) {
            StringBuffer sb = new StringBuffer();
            DOMUtils.getStringValue((TypedNode) resnodes.get(0), sb);
            resvalue = new Long(sb.length());
        } else if (resnodes != null && resnodes.size() > 1) {
            throw new XQueryException("Evaluation : incorrect argument for function string-length");
        }
        Integer rettype = (Integer) returnType.peek();
        if (rettype == NODE_TYPE) {
            ArrayList nodes = new ArrayList(1);
            TypedValue text = docimpl.createTypedValue(resvalue, this.schemamanager.getIntegerType());
            nodes.add(text);
            setOutput(nodes);
            return;
        } else {
            setOutput(resvalue);
            return;
        }
        // clear();
    }

    /*
     * 7.4.6 fn:substring fn:substring( $sourceString as xs:string?, $startingLoc as xs:double) as xs:string?
     * 
     * fn:substring( $sourceString as xs:string?, $startingLoc as xs:double, $length as xs:double) as xs:string?
     * 
     * If the value of $sourceString is the empty sequence, the empty sequence is returned.
     * 
     * Otherwise, returns the portion of the value of $sourceString beginning at the position indicated by the value of $startingLoc and continuing for the number of characters indicated by the value of $length. More specifically, returns the characters in $sourceString whose position $p obeys:
     * 
     * fn:round($startingLoc) <= $p < fn:round($startingLoc) + fn:round($length)
     * 
     * In the above computation, the rules for op:numeric-less-than() and op:numeric-greater-than() apply.
     * 
     * If $startingLoc is zero or negative, the substring includes characters from the beginning of the $sourceString.
     * 
     * If $length is not specified, the substring includes characters to the end of $sourceString.
     * 
     * If $length is greater than the number of characters in the value of $sourceString following $startingLoc, the substring includes characters to the end of $sourceString.
     * 
     * The first character of a string is located at position 1, not position 0.
     * 
     * Note:
     * 
     * The position and length given in the second and (optional) third argument relate to the number of XML characters in the string (or equivalently, the number of Unicode code points). Some implementations may represent a code point above xFFFF using two 16-bit values known as a surrogate pair. A surrogate pair counts as one character, not two.
     */
    public void visit(FunctionSUBSTRING arg) throws XQueryException {
        boolean hasLength = (arg.getArguments().size() == 3) ? true : false;
        String sourceString = null;
        returnType.push(UNKNOWN_TYPE);
        arg.getArgument(0).accept(this);
        if (resvalue != null) {
            if (resvalue instanceof String)
                sourceString = (String) resvalue;
            else
                sourceString = resvalue.toString();
        } else if (resnodes != null && resnodes.size() == 1) {
            StringBuffer sb = new StringBuffer();
            DOMUtils.getStringValue((TypedNode) resnodes.get(0), sb);
            sourceString = sb.toString();
        } else if (resnodes != null && resnodes.size() > 1) {
            throw new XQueryException("Evaluation : incorrect first argument for function substring");
        }
        if (sourceString == null) {
            clear();
            return;
        }
        double startingLoc = 0;
        arg.getArgument(1).accept(this);
        if (resvalue != null) { // do nothing
        } else if (resnodes != null && resnodes.size() == 1) {
            TypedNode tmpnode = (TypedNode) ((TypedNode) resnodes.get(0)).getFirstChild();
            if (tmpnode == null)
                tmpnode = (TypedNode) resnodes.get(0);
            resvalue = (Comparable) tmpnode.getTypedValue();
        } else
            throw new XQueryException("Evaluation : incorrect second argument for function substring");
        if (resvalue instanceof Number) {
            startingLoc = ((Number) resvalue).doubleValue();
        } else if (resvalue instanceof String) {
            try {
                startingLoc = Double.parseDouble((String) resvalue);
            } catch (NumberFormatException e) {
                // do nothing
            }
        } else
            throw new XQueryException("Evaluation : incorrect second argument for function substring");
        if (startingLoc < 1)
            startingLoc = 1;
        double length = sourceString.length();
        if (hasLength) {
            arg.getArgument(2).accept(this);
            if (resvalue != null) { // do nothing
            } else if (resnodes != null && resnodes.size() == 1) {
                TypedNode tmpnode = (TypedNode) ((TypedNode) resnodes.get(0)).getFirstChild();
                if (tmpnode == null)
                    tmpnode = (TypedNode) resnodes.get(0);
                resvalue = (Comparable) tmpnode.getTypedValue();
            } else
                throw new XQueryException("Evaluation : incorrect third argument for function substring");
            if (resvalue instanceof Number) {
                length = ((Number) resvalue).doubleValue();
            } else if (resvalue instanceof String) {
                try {
                    length = Double.parseDouble((String) resvalue);
                } catch (NumberFormatException e) {
                    // do nothing
                }
            } else
                throw new XQueryException("Evaluation : incorrect third argument for function substring");
            if (startingLoc + length > sourceString.length())
                hasLength = false;
        }
        returnType.pop();
        if (hasLength)
            resvalue = sourceString.substring((int) Math.round(startingLoc) - 1, (int) (Math.round(startingLoc) + Math.round(length)) - 1);
        else
            resvalue = sourceString.substring((int) Math.round(startingLoc) - 1);
        Integer rettype = (Integer) returnType.peek();
        if (rettype == NODE_TYPE) {
            ArrayList nodes = new ArrayList(1);
            TypedValue text = docimpl.createTypedValue(resvalue, this.schemamanager.getStringType());
            nodes.add(text);
            setOutput(nodes);
            return;
        } else {
            setOutput(resvalue);
            return;
        }
        // clear();
    }

    // -----------------------------------------------------------
    // MATHEMATICAL FUNCTIONS
    // -----------------------------------------------------------
    public void visit(FunctionABS arg) throws XQueryException {
        returnType.push(VALUE_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        // if (resvalue != null) { // do nothing
        // } else if (resnodes == null || resnodes.isEmpty()) {
        // resvalue = null;
        // } else if (resnodes != null && resnodes.size() == 1) {
        // TypedNode tmpnode = (TypedNode) ((TypedNode) resnodes.get(0)).getFirstChild();
        // if (tmpnode == null)
        // tmpnode = (TypedNode) resnodes.get(0);
        // resvalue = (Comparable) tmpnode.getTypedValue();
        // if (!(resvalue instanceof Number))
        // throw new XQueryException("Could not apply function abs");
        // } else
        // throw new XQueryException("Could not apply function abs");

        if (resvalue != null) {
            if (resvalue instanceof Long) {
                resvalue = new Long(Math.abs(((Long) resvalue).longValue()));
                setOutput(resvalue, schemamanager.getIntegerType());
            } else if (resvalue instanceof BigDecimal) {
                resvalue = ((BigDecimal) resvalue).abs();
                setOutput(resvalue, schemamanager.getDecimalType());
            } else if (resvalue instanceof Float) {
                resvalue = new Float(Math.abs(((Float) resvalue).floatValue()));
                setOutput(resvalue, schemamanager.getFloatType());
            } else if (resvalue instanceof Double) {
                resvalue = new Double(Math.abs(((Double) resvalue).doubleValue()));
                setOutput(resvalue, schemamanager.getDoubleType());
            } else
                throw new XQueryException("Could not apply function abs");
            return;
        }

        setOutput(resvalue);
    }

    public void visit(FunctionCEILING arg) throws XQueryException {
        returnType.push(UNKNOWN_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        if (resvalue != null) { // do nothing
        } else if (resnodes != null && resnodes.size() == 1) {
            TypedNode tmpnode = (TypedNode) ((TypedNode) resnodes.get(0)).getFirstChild();
            if (tmpnode == null)
                tmpnode = (TypedNode) resnodes.get(0);
            resvalue = (Comparable) tmpnode.getTypedValue();
            if (!(resvalue instanceof Number))
                throw new XQueryException("Could not apply function ceiling");
        } else
            throw new XQueryException("Could not apply function ceiling");

        if (resvalue instanceof Long) { // do nothing
        } else if (resvalue instanceof BigDecimal)
            resvalue = ((BigDecimal) resvalue).setScale(0, BigDecimal.ROUND_CEILING);
        else if (resvalue instanceof Float)
            resvalue = new Float(Math.ceil(((Float) resvalue).floatValue()));
        else if (resvalue instanceof Double)
            resvalue = new Double(Math.ceil(((Double) resvalue).doubleValue()));

        Integer rettype = (Integer) returnType.peek();
        if (rettype == VALUE_TYPE) {
            setOutput(resvalue);
            return;
        } else { // (rettype == NODE_TYPE) {
            ArrayList nodes = new ArrayList(1);
            TypedValue text = docimpl.createTypedValue(resvalue, this.getSimpleType(resvalue, schemamanager.getDoubleType()));
            nodes.add(text);
            setOutput(nodes);
            return;
        }
        // clear();
    }

    public void visit(FunctionFLOOR arg) throws XQueryException {
        returnType.push(UNKNOWN_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        if (resvalue != null) { // do nothing
        } else if (resnodes != null && resnodes.size() == 1) {
            TypedNode tmpnode = (TypedNode) ((TypedNode) resnodes.get(0)).getFirstChild();
            if (tmpnode == null)
                tmpnode = (TypedNode) resnodes.get(0);
            resvalue = (Comparable) tmpnode.getTypedValue();
            if (!(resvalue instanceof Number))
                throw new XQueryException("Could not apply function floor");
        } else
            throw new XQueryException("Could not apply function floor");

        if (resvalue instanceof Long) { // do nothing
        } else if (resvalue instanceof BigDecimal)
            resvalue = ((BigDecimal) resvalue).setScale(0, BigDecimal.ROUND_FLOOR);
        else if (resvalue instanceof Float)
            resvalue = new Float(Math.floor(((Float) resvalue).floatValue()));
        else if (resvalue instanceof Double)
            resvalue = new Double(Math.floor(((Double) resvalue).doubleValue()));

        Integer rettype = (Integer) returnType.peek();
        if (rettype == VALUE_TYPE) {
            setOutput(resvalue);
            return;
        } else { // (rettype == NODE_TYPE) {
            ArrayList nodes = new ArrayList(1);
            TypedValue text = docimpl.createTypedValue(resvalue, this.getSimpleType(resvalue, schemamanager.getDoubleType()));
            nodes.add(text);
            setOutput(nodes);
            return;
        }
        // clear();
    }

    public void visit(FunctionROUND arg) throws XQueryException {
        returnType.push(UNKNOWN_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        if (resvalue != null) { // do nothing
        } else if (resnodes != null && resnodes.size() == 1) {
            TypedNode tmpnode = (TypedNode) ((TypedNode) resnodes.get(0)).getFirstChild();
            if (tmpnode == null)
                tmpnode = (TypedNode) resnodes.get(0);
            resvalue = (Comparable) tmpnode.getTypedValue();
            if (!(resvalue instanceof Number))
                throw new XQueryException("Could not apply function round");
        } else
            throw new XQueryException("Could not apply function round");

        if (resvalue instanceof Long) { // do nothing
        } else if (resvalue instanceof BigDecimal)
            resvalue = ((BigDecimal) resvalue).setScale(0, BigDecimal.ROUND_UP);
        else if (resvalue instanceof Float)
            resvalue = new BigDecimal(Math.round(((Float) resvalue).floatValue()));
        else if (resvalue instanceof Double)
            resvalue = new BigDecimal(Math.round(((Double) resvalue).doubleValue()));

        Integer rettype = (Integer) returnType.peek();
        if (rettype == VALUE_TYPE) {
            setOutput(resvalue);
            return;
        } else { // (rettype == NODE_TYPE) {
            ArrayList nodes = new ArrayList(1);
            TypedValue text = docimpl.createTypedValue(resvalue, this.getSimpleType(resvalue, schemamanager.getDoubleType()));
            nodes.add(text);
            setOutput(nodes);
            return;
        }
        // clear();
    }

    /*
     * 14.1.4 fn:number fn:number() as xs:double fn:number($srcval as item?) as xs:double Returns the value indicated by $srcval or, if $srcval is not specified, the context node, converted to an xs:double. If there is no context node (that is, if there is no context item, or if the context item is not a node), NaN is returned.
     * 
     * If $srcval is the empty sequence, returns the xs:double value NaN.
     * 
     * If $srcval is atomic, returns the value obtained by converting it to xs:double following the rules of 17.8 Casting to numeric types.
     * 
     * If $srcval is a node with an atomic type, returns that value converted to xs:double following the rules of 17.8 Casting to numeric types.
     * 
     * Otherwise, converts $srcval to an xs:string as if by a call to the fn:string() function and then converts the result to an xs:double following the rules of 17.8 Casting to numeric types.
     * 
     * If the conversion to xs:double fails because the lexical representation is not a valid lexical representation of a numeric simple type as defined in [XML Schema Part 2: Datatypes], returns the xs:double value NaN.
     */
    public void visit(FunctionNUMBER arg) throws XQueryException {
        returnType.push(UNKNOWN_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        if (resvalue != null) { // do nothing
        } else if (resnodes != null && resnodes.size() == 1) {
            TypedNode tmpnode = (TypedNode) ((TypedNode) resnodes.get(0)).getFirstChild();
            if (tmpnode == null)
                tmpnode = (TypedNode) resnodes.get(0);
            resvalue = (Comparable) tmpnode.getTypedValue();
            if (resvalue instanceof Number) {
                resvalue = new Double(((Number) resvalue).doubleValue());
            } else if (resvalue instanceof String) {
                try {
                    resvalue = new Double(Double.parseDouble((String) resvalue));
                } catch (NumberFormatException e) {
                    // do nothing
                }
            }
        } else if (resnodes != null && resnodes.size() > 1) {
            throw new XQueryException("Evaluation : incorrect argument for function number");
        }
        if (resvalue == null) {
            resvalue = new Double(Double.NaN);
        }
        Integer rettype = (Integer) returnType.peek();
        if (rettype == NODE_TYPE) {
            ArrayList nodes = new ArrayList(1);
            TypedValue text = docimpl.createTypedValue(resvalue, this.schemamanager.getDoubleType());
            nodes.add(text);
            setOutput(nodes);
            return;
        } else {
            setOutput(resvalue);
            return;
        }
        // clear();
    }

    // -----------------------------------------------------------
    // CAST/CONSTRUCTOR FUNCTIONS
    // -----------------------------------------------------------
    public void visit(FunctionDOUBLE arg) throws XQueryException {
        returnType.push(UNKNOWN_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        if (resvalue != null) { // do nothing
        } else if (resnodes != null && resnodes.size() == 1) {
            TypedNode tmpnode = (TypedNode) ((TypedNode) resnodes.get(0)).getFirstChild();
            if (tmpnode == null)
                tmpnode = (TypedNode) resnodes.get(0);
            resvalue = (Comparable) tmpnode.getTypedValue();
        } else
            throw new XQueryException("Could not create double");

        if (resvalue instanceof Double) { // do nothing
        } else if (resvalue instanceof Number) {
            resvalue = new Double(((Number) resvalue).doubleValue());
        } else if (resvalue instanceof String) {
            try {
                resvalue = new Double(Double.parseDouble((String) resvalue));
            } catch (NumberFormatException e) {
                // do nothing
            }
        } else
            throw new XQueryException("Could not create double");
        Integer rettype = (Integer) returnType.peek();
        if (rettype == VALUE_TYPE) {
            setOutput(resvalue);
            return;
        } else { // (rettype == NODE_TYPE) {
            ArrayList nodes = new ArrayList(1);
            TypedValue text = docimpl.createTypedValue(resvalue, this.schemamanager.getDoubleType());
            nodes.add(text);
            setOutput(nodes);
            return;
        }
        // clear();
    }

    public void visit(FunctionINTEGER arg) throws XQueryException {
        returnType.push(UNKNOWN_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        if (resvalue != null) { // do nothing
        } else if (resnodes != null && resnodes.size() == 1) {
            TypedNode tmpnode = (TypedNode) ((TypedNode) resnodes.get(0)).getFirstChild();
            if (tmpnode == null)
                tmpnode = (TypedNode) resnodes.get(0);
            resvalue = (Comparable) tmpnode.getTypedValue();
        } else
            throw new XQueryException("Could not create integer");

        if (resvalue instanceof Long) { // do nothing
        } else if (resvalue instanceof Number) {
            resvalue = new Long(((Number) resvalue).longValue());
        } else if (resvalue instanceof String) {
            try {
                resvalue = new Long((String) resvalue);
            } catch (NumberFormatException e) {
                // do nothing
            }
        } else
            throw new XQueryException("Could not create integer");
        Integer rettype = (Integer) returnType.peek();
        if (rettype == VALUE_TYPE) {
            setOutput(resvalue);
            return;
        } else { // (rettype == NODE_TYPE) {
            ArrayList nodes = new ArrayList(1);
            TypedValue text = docimpl.createTypedValue(resvalue, this.schemamanager.getIntegerType());
            nodes.add(text);
            setOutput(nodes);
            return;
        }
        // clear();
    }

    public void visit(FunctionDECIMAL arg) throws XQueryException {
        returnType.push(UNKNOWN_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        if (resvalue != null) { // do nothing
        } else if (resnodes != null && resnodes.size() == 1) {
            TypedNode tmpnode = (TypedNode) ((TypedNode) resnodes.get(0)).getFirstChild();
            if (tmpnode == null)
                tmpnode = (TypedNode) resnodes.get(0);
            resvalue = (Comparable) tmpnode.getTypedValue();
        } else
            throw new XQueryException("Could not create decimal");

        if (resvalue instanceof BigDecimal) { // do nothing
        } else if (resvalue instanceof Number) {
            resvalue = new BigDecimal(((Number) resvalue).doubleValue());
        } else if (resvalue instanceof String) {
            try {
                resvalue = new BigDecimal((String) resvalue);
            } catch (NumberFormatException e) {
                // do nothing
            }
        } else
            throw new XQueryException("Could not create decimal");
        Integer rettype = (Integer) returnType.peek();
        if (rettype == VALUE_TYPE) {
            setOutput(resvalue);
            return;
        } else { // (rettype == NODE_TYPE) {
            ArrayList nodes = new ArrayList(1);
            TypedValue text = docimpl.createTypedValue(resvalue, this.schemamanager.getDecimalType());
            nodes.add(text);
            setOutput(nodes);
            return;
        }
        // clear();
    }

    // -----------------------------------------------------------
    // OTHER FUNCTIONS
    // -----------------------------------------------------------

    public void visit(FunctionBASE_URI arg) throws XQueryException {
        // TODO this is the default implementation, no base uri is ever returned!!!
        setOutput(null, this.schemamanager.getAnyURIType());
    }

    public void visit(FunctionDATA arg) throws XQueryException {
        returnType.push(UNKNOWN_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        Integer rettype = (Integer) returnType.peek();
        if (resvalue != null) {
            if (rettype == VALUE_TYPE) {
                setOutput(resvalue);
                return;
            } else { // (rettype == NODE_TYPE) {
                ArrayList nodes = new ArrayList(1);
                TypedValue text = docimpl.createTypedValue(resvalue, this.getSimpleType(resvalue, schemamanager.getAnySimpleType()));
                nodes.add(text);
                setOutput(nodes);
                return;
            }
        } else if (resnodes != null) {
            ArrayList tmplist = new ArrayList(resnodes.size());
            for (int i = 0; i < resnodes.size(); i++) {
                TypedNode tmpnode = (TypedNode) ((TypedNode) resnodes.get(i)).getFirstChild();
                // change 02/09/2003
                if (tmpnode == null || tmpnode.getTypedValue() == null)
                    tmpnode = (TypedNode) resnodes.get(i);
                Comparable comp = (Comparable) tmpnode.getTypedValue();
                if (comp != null)
                    tmplist.add(docimpl.createTypedValue(comp, this.getSimpleType(comp, schemamanager.getAnySimpleType())));
                // else
                // throw new XQueryException("Could not apply function data");
            }
            if (rettype == VALUE_TYPE) {
                if (tmplist.size() == 1) {
                    TypedNode tmpnode = (TypedNode) ((TypedNode) tmplist.get(0)).getFirstChild();
                    if (tmpnode == null)
                        tmpnode = (TypedNode) tmplist.get(0);
                    setOutput((Comparable) tmpnode.getTypedValue());
                    return;
                }
            } else if (rettype == BOOLEAN_TYPE) {
                if (tmplist.size() > 0)
                    setOutput(true);
                else
                    setOutput(false);
                return;
            } else { // (rettype == NODE_TYPE) {
                setOutput(tmplist);
                return;
            }
        }
        // else
        // throw new XQueryException("Could not apply function data");
        // clear();
    }

    public void visit(FunctionEXISTS arg) throws XQueryException {
        returnType.push(NODE_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        setOutput(resnodes != null && !resnodes.isEmpty());
    }

    public void visit(FunctionEMPTY arg) throws XQueryException {
        // if (buffertuple != null && !buffertuple.isEmpty()) {
        // setOutput(false);
        // return;
        // }
        returnType.push(NODE_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        boolean tmpVerdict = resnodes == null || resnodes.isEmpty();

        setOutput(tmpVerdict);

        // setOutput(resnodes == null || resnodes.isEmpty());
        //        
        //        
        // Integer rettype = (Integer) returnType.peek();
        //        
        // if (rettype == VALUE_TYPE) {
        // throw new EvaluatorException("Could not transform boolean into
        // Comparable");
        // } else if (rettype == NODE_TYPE) {
        // ArrayList nodes = null;
        // if (resvalue != null) {
        // nodes = new ArrayList(1);
        // TypedValue text = docimpl.createTypedValue(resvalue,
        // getSimpleType(resvalue, schemamanager.getDoubleType()));
        // nodes.add(text);
        // }
        // setResultNodes(nodes);
        // return;
        // } else if (rettype == )

    }

    public void visit(FunctionNOT arg) throws XQueryException {
        returnType.push(BOOLEAN_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        setOutput(!resverdict);
    }

    public void visit(FunctionDISTINCT_VALUES arg) throws XQueryException {
        returnType.push(NODE_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        if (resnodes != null) {
            // cloning must be done otherwise context tuple could be changed
            resnodes = (ArrayList) resnodes.clone();
            for (int i = 0; i < resnodes.size(); i++) {
                TypedNode nodei = (TypedNode) resnodes.get(i);
                if (nodei.getTypedValue() == null) {
                    // apply data function
                    nodei = (TypedNode) nodei.getFirstChild();
                    if (nodei == null || nodei.getTypedValue() == null) {
                        resnodes.remove(i);
                        i--;
                        continue;
                    }
                    resnodes.set(i, nodei);
                }
                for (int j = 0; j < i; j++) {
                    TypedNode nodej = (TypedNode) resnodes.get(j);
                    if (nodej.getTypedValue() == null) {
                        // apply data function
                        nodej = (TypedNode) nodej.getFirstChild();
                        if (nodej == null || nodej.getTypedValue() == null) {
                            resnodes.remove(j);
                            continue;
                        }
                        resnodes.set(j, nodej);
                    }
                    if (DOMUtils.compareTree(nodei, nodej)) {
                        resnodes.remove(i);
                        i--;
                        break;
                    }
                }
            }
        }
        setOutput(resnodes);
    }

    public void visit(FunctionDEEP_EQUAL arg) throws XQueryException {
        resnodes = null;
        XQueryExpression arg0 = arg.getArgument(0);
        XQueryExpression arg1 = arg.getArgument(1);
        if (arg0 instanceof Value && arg1 instanceof Value)
            processComparison(arg0, arg1, null, null, null, null, Constants.EQ_COMPOP);
        else {
            if (arg0 instanceof Value) {
                if (tuple != null) {
                    ArrayList nodes1 = tuple.getNodesAtIndex(tuple.getParent().getPathIndex(arg1.getStringValue()));
                    if (nodes1 != null && nodes1.size() == 1) {
                        processComparison(arg0, arg1, (TypedNode) null, (TypedNode) nodes1.get(0), null, null, Constants.EQ_COMPOP);
                    }
                } else if (tuplelist != null) {
                    for (int i = 0; i < tuplelist.size(); i++) {
                        Tuple tuplei = (Tuple) tuplelist.get(i);
                        ArrayList nodes1 = tuplei.getNodesAtIndex(tuplei.getParent().getPathIndex(arg1.getStringValue()));
                        if (nodes1 != null && nodes1.size() == 1) {
                            processComparison(arg0, arg1, (TypedNode) null, (TypedNode) nodes1.get(0), null, null, Constants.EQ_COMPOP);
                            break;
                        }
                    }
                }
            } else if (arg1 instanceof Value) {
                if (tuple != null) {
                    ArrayList nodes0 = tuple.getNodesAtIndex(tuple.getParent().getPathIndex(arg0.getStringValue()));
                    if (nodes0 != null && nodes0.size() == 1) {
                        processComparison(arg0, arg1, (TypedNode) nodes0.get(0), (TypedNode) null, null, null, Constants.EQ_COMPOP);
                    }
                } else if (tuplelist != null) {
                    for (int i = 0; i < tuplelist.size(); i++) {
                        Tuple tuplei = (Tuple) tuplelist.get(i);
                        ArrayList nodes0 = tuplei.getNodesAtIndex(tuplei.getParent().getPathIndex(arg0.getStringValue()));
                        if (nodes0 != null && nodes0.size() == 1) {
                            processComparison(arg0, arg1, (TypedNode) nodes0.get(0), (TypedNode) null, null, null, Constants.EQ_COMPOP);
                            break;
                        }
                    }
                }
            } else {
                if (this.tuple != null) {
                    ArrayList nodes0 = tuple.getNodesAtIndex(tuple.getParent().getPathIndex(arg0.getStringValue()));
                    ArrayList nodes1 = tuple.getNodesAtIndex(tuple.getParent().getPathIndex(arg1.getStringValue()));
                    setOutput(DOMUtils.compareTree(nodes0, nodes1));
                } else if (tuplelist != null) {
                    ArrayList nodes0 = null;
                    ArrayList nodes1 = null;
                    for (int i = 0; i < tuplelist.size(); i++) {
                        Tuple tmptuple = (Tuple) tuplelist.get(i);
                        if (nodes0 == null)
                            nodes0 = tmptuple.getNodesAtIndex(tmptuple.getParent().getPathIndex(arg0.getStringValue()));
                        if (nodes1 == null)
                            nodes1 = tmptuple.getNodesAtIndex(tmptuple.getParent().getPathIndex(arg1.getStringValue()));
                        if (nodes0 != null && nodes1 != null)
                            break;
                    }
                    setOutput(DOMUtils.compareTree(nodes0, nodes1));
                    // } else if (buffertuple != null) {
                    // throw new XQueryException("FunctionVALUE_EQUALS on
                    // BufferTuple not yet implemented");
                }
            }
        }
    }

    // -----------------------------------------------------------
    // QNAME FUNCTIONS
    // -----------------------------------------------------------

    public void visit(org.xquark.xquery.parser.primitivefunctions.fnfunctions.FunctionQNAME arg) throws XQueryException {
        returnType.push(VALUE_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        if (resvalue == null)
            throw new XQueryException("Could not evaluate " + arg.getArgument(0));
        String prefix = resvalue.toString();
        String namespace = null;
        if (arg.getArguments().size() > 1) {
            if (prefix != null && !prefix.startsWith("{")) {
                namespace = context.getNamespaceURI(prefix);
                if (namespace == null)
                    throw new XQueryException("Could not find prefix : " + prefix);
            }
            returnType.push(VALUE_TYPE);
            arg.getArgument(1).accept(this);
            returnType.pop();
            if (resvalue == null)
                throw new XQueryException("Could not evaluate " + arg.getArgument(1));
            if (prefix != null) {
                if (!prefix.startsWith("{"))
                    setOutput(prefix+":"+resvalue);
                else
                    setOutput(prefix+resvalue);
            } else 
                setOutput(resvalue);
           
        } else {
            String qname = prefix;
            prefix = null;
            if (!qname.startsWith("{")) {
                int index = qname.indexOf(":");
                if (index != -1) {
                    prefix = qname.substring(0, index);
                    qname = qname.substring(index + 1);
                    namespace = context.getNamespaceURI(prefix);
                    if (namespace == null)
                        throw new XQueryException("Could not find prefix : " + prefix);
                }
            }
            setOutput((prefix == null) ? qname : prefix + ":" + qname);
        }
    }

    public void visit(org.xquark.xquery.parser.primitivefunctions.xsfunctions.FunctionQNAME arg) throws XQueryException {
        returnType.push(VALUE_TYPE);
        arg.getArgument(0).accept(this);
        returnType.pop();
        if (resvalue == null) {
            setOutput((Comparable)null);
            return;
        }
        String qname = resvalue.toString();
        String prefix = null;
        if (!qname.startsWith("{")) {
            int index = qname.indexOf(":");
            if (index != -1) {
                prefix = qname.substring(0, index);
                qname = qname.substring(index + 1);
                String namespace = context.getNamespaceURI(prefix);
                if (namespace == null)
                    throw new XQueryException("Could not find prefix : " + prefix);
            }
        }
        setOutput((prefix == null) ? qname : prefix + ":" + qname);
    }

    // -----------------------------------------------------------
    // PRIVATE METHODS
    // -----------------------------------------------------------

    private SimpleType getSimpleType(Comparable obj) {
        return getSimpleType(obj, this.schemamanager.getAnySimpleType());
    }

    private SimpleType getSimpleType(Comparable obj, SimpleType defaultType) {
        if (obj instanceof Long)
            return this.schemamanager.getIntegerType();
        else if (obj instanceof BigDecimal)
            return this.schemamanager.getDecimalType();
        else if (obj instanceof Float)
            return this.schemamanager.getFloatType();
        else if (obj instanceof Double)
            return this.schemamanager.getDoubleType();
        else if (obj instanceof String)
            return this.schemamanager.getStringType();
        else if (obj instanceof DateTime)
            return this.schemamanager.getDateTimeType();
        return defaultType;
    }

    private void processComparison(XQueryExpression exp1, XQueryExpression exp2, TypedNode tree1, TypedNode tree2, Comparable value1, Comparable value2, int operator) throws EvaluatorException {
        // Comparable value1 = null;
        boolean doublecompare = false;
        double value1double = 0;
        if (value1 == null) {
            if (exp1 instanceof ValueInteger) {
                value1 = new Long(((ValueInteger) exp1).getIntValue().longValue());
            } else if (exp1 instanceof ValueDecimal) {
                value1 = ((ValueDecimal) exp1).getDecimalValue();
            } else if (exp1 instanceof ValueFloat) {
                value1 = ((ValueFloat) exp1).getFloatValue();
            } else if (exp1 instanceof ValueDouble) {
                value1 = ((ValueDouble) exp1).getDoubleValue();
            } else if (exp1 instanceof ValueString) {
                value1 = ((ValueString) exp1).getValue();
            } else if (tree1 != null) {
                if (tree1.getTypedValue() != null)
                    value1 = (Comparable) tree1.getTypedValue();
                else
                    value1 = (Comparable) DOMUtils.getTextLeaf(tree1);
            }
        }
        if (value1 != null && !(value1 instanceof String) && !(value1 instanceof Timestamp) && !(value1 instanceof DateTime) && !(value1 instanceof Duration)) {
            value1double = ((Number) value1).doubleValue();
            doublecompare = true;
        }
        // Comparable value2 = null;
        double value2double = 0;
        if (value2 == null) {
            if (exp2 instanceof ValueInteger) {
                value2 = new Long(((ValueInteger) exp2).getIntValue().longValue());
            } else if (exp2 instanceof ValueDecimal) {
                value2 = ((ValueDecimal) exp2).getDecimalValue();
            } else if (exp2 instanceof ValueFloat) {
                value2 = ((ValueFloat) exp2).getFloatValue();
            } else if (exp2 instanceof ValueDouble) {
                value2 = ((ValueDouble) exp2).getDoubleValue();
            } else if (exp2 instanceof ValueString)
                value2 = ((ValueString) exp2).getValue();
            else if (tree2 != null) {
                if (tree2.getTypedValue() != null)
                    value2 = (Comparable) tree2.getTypedValue();
                else
                    value2 = (Comparable) DOMUtils.getTextLeaf(tree2);
            }
        }
        if (value2 != null && !(value2 instanceof String) && !(value2 instanceof Timestamp) && !(value2 instanceof DateTime) && !(value2 instanceof Duration)) {
            value2double = ((Number) value2).doubleValue();
        } else
            doublecompare = false;
        boolean stringcomp = false;
        if (value1 == null) {
            if (value2 != null) {
                setOutput(false);
                return;
            }
        } else {
            if (value2 == null) {
                setOutput(false);
                return;
            } else {
                if (value1 instanceof String) {
                    if (!(value2 instanceof String)) {
                        value2 = value2.toString();
                        if (value2 == null) {
                            setOutput(false);
                            return;
                        }
                        stringcomp = true;
                    }
                } else if (value2 instanceof String) {
                    value1 = value1.toString();
                    if (value1 == null) {
                        setOutput(false);
                        return;
                    }
                    stringcomp = true;
                }
            }
        }
        switch (operator) {
        case Constants.LT_VALUECOMP:
        case Constants.LT_COMPOP:
            if ((value1 == null) && (value2 == null))
                setOutput(false);
            else if (stringcomp || !doublecompare)
                setOutput((DOMUtils.compareObjects(value1, value2) < 0));
            else
                setOutput((value1double < value2double));
            return;
        case Constants.GT_VALUECOMP:
        case Constants.GT_COMPOP:
            if ((value1 == null) && (value2 == null))
                setOutput(false);
            else if (stringcomp || !doublecompare)
                setOutput((DOMUtils.compareObjects(value1, value2) > 0));
            else
                setOutput((value1double > value2double));
            return;
        case Constants.GE_VALUECOMP:
        case Constants.GEQ_COMPOP:
            if ((value1 == null) && (value2 == null))
                setOutput(true);
            else if (stringcomp || !doublecompare)
                setOutput((DOMUtils.compareObjects(value1, value2) >= 0));
            else
                setOutput((value1double >= value2double));
            return;
        case Constants.LE_VALUECOMP:
        case Constants.LEQ_COMPOP:
            if ((value1 == null) && (value2 == null))
                setOutput(true);
            else if (stringcomp || !doublecompare)
                setOutput((DOMUtils.compareObjects(value1, value2) <= 0));
            else
                setOutput((value1double <= value2double));
            return;
        case Constants.EQ_VALUECOMP:
        case Constants.EQ_COMPOP:
            if ((value1 == null) && (value2 == null))
                setOutput(true);
            else if (stringcomp || !doublecompare)
                setOutput((DOMUtils.compareObjects(value1, value2) == 0));
            else
                setOutput((value1double == value2double));
            return;
        case Constants.NE_VALUECOMP:
        case Constants.NEQ_COMPOP:
            if ((value1 == null) && (value2 == null))
                setOutput(false);
            else if (stringcomp || !doublecompare)
                setOutput((DOMUtils.compareObjects(value1, value2) != 0));
            else
                setOutput((value1double != value2double));
            return;
            /*
             * case Constants.EQQ_COMPOP : setVerdict(Tuple.compareTree(tree1, tree2)); return;
             */
        }
        clear();
    }

    private Number processOperation(XQueryExpression exp1, XQueryExpression exp2, TypedNode tree1, TypedNode tree2, Number value1, Number value2, int operator) throws XQueryException {
        if (value1 == null) {
            if (exp1 instanceof ValueInteger)
                value1 = new Long(((ValueInteger) exp1).getIntValue().longValue());
            else if (exp1 instanceof ValueDecimal)
                value1 = ((ValueDecimal) exp1).getDecimalValue();
            else if (exp1 instanceof ValueFloat)
                value1 = ((ValueFloat) exp1).getFloatValue();
            else if (exp1 instanceof ValueDouble)
                value1 = ((ValueDouble) exp1).getDoubleValue();
            else if (tree1 != null)
                if (tree1.getTypedValue() != null)
                    value1 = (Number) tree1.getTypedValue();
                else
                    value1 = (Number) DOMUtils.getTextLeaf(tree1);
        }
        if (value2 == null) {
            if (exp2 instanceof ValueInteger)
                value2 = new Long(((ValueInteger) exp2).getIntValue().longValue());
            else if (exp2 instanceof ValueDecimal)
                value2 = ((ValueDecimal) exp2).getDecimalValue();
            else if (exp2 instanceof ValueFloat)
                value2 = ((ValueFloat) exp2).getFloatValue();
            else if (exp2 instanceof ValueDouble)
                value2 = ((ValueDouble) exp2).getDoubleValue();
            else if (tree2 != null)
                if (tree2.getTypedValue() != null)
                    value2 = (Number) tree2.getTypedValue();
                else
                    value2 = (Number) DOMUtils.getTextLeaf(tree2);
        }

        boolean integerOp = false;
        boolean decimalOp = false;
        boolean floatOp = false;
        if (value1 instanceof Long) {
            if (value2 instanceof Long)
                integerOp = true;
            else if (value2 instanceof BigDecimal) {
                value1 = new BigDecimal(value1.longValue());
                decimalOp = true;
            }
        } else if (value1 instanceof BigDecimal) {
            if (value2 instanceof BigDecimal)
                decimalOp = true;
            else if (value2 instanceof Long) {
                value2 = new BigDecimal(value2.longValue());
                decimalOp = true;
            }
        } else if (value1 instanceof Float) {
            if (value2 instanceof Float)
                floatOp = true;
            if (!(value2 instanceof Double)) {
                value2 = new Float(value2.doubleValue());
                floatOp = true;
            }
        }

        switch (operator) {
        case Constants.DIVIDE_ARITHMETICS:
            // if (integerOp) {
            // try {
            // return value1.intValue() / value2.intValue();
            // } catch (ArithmeticException ae) {
            // throw new XQueryException("Division by zero");
            // }
            // } else
            if (decimalOp) {
                try {
                    return ((BigDecimal) value1).divide((BigDecimal) value2, 38, BigDecimal.ROUND_UP);
                } catch (ArithmeticException ae) {
                    throw new XQueryException("Division by zero");
                }
            } else if (floatOp)
                return new Float(value1.floatValue() / value2.floatValue());
            return new Double(value1.doubleValue() / value2.doubleValue());
        case Constants.MINUS_ARITHMETICS:
            if (integerOp)
                return new Long(value1.longValue() - value2.longValue());
            else if (decimalOp)
                return ((BigDecimal) value1).subtract((BigDecimal) value2);
            else if (floatOp)
                return new Float(value1.floatValue() - value2.floatValue());
            return new Double(value1.doubleValue() - value2.doubleValue());
        case Constants.MODULO_ARITHMETICS:
            if (integerOp)
                return new Long(value1.longValue() % value2.longValue());
            else if (floatOp)
                return new Float(value1.floatValue() % value2.floatValue());
            return new Double(value1.doubleValue() % value2.doubleValue());
        case Constants.MULTIPLY_ARITHMETICS:
            if (integerOp)
                return new Long(value1.longValue() * value2.longValue());
            else if (decimalOp)
                return ((BigDecimal) value1).multiply((BigDecimal) value2);
            else if (floatOp)
                return new Float(value1.floatValue() * value2.floatValue());
            return new Double(value1.doubleValue() * value2.doubleValue());
        case Constants.PLUS_ARITHMETICS:
            if (integerOp)
                return new Long(value1.longValue() + value2.longValue());
            else if (decimalOp)
                return ((BigDecimal) value1).add((BigDecimal) value2);
            else if (floatOp)
                return new Float(value1.floatValue() + value2.floatValue());
            return new Double(value1.doubleValue() + value2.doubleValue());
        }
        throw new XQueryException("Could not process operation");
    }

}