package org.xquark.xquery.typing;

import java.util.Iterator;
import java.util.List;

import javax.xml.namespace.QName;

import org.xml.sax.SAXException;
import org.xquark.schema.ChoiceModelGroup;
import org.xquark.schema.ComplexType;
import org.xquark.schema.ContentModel;
import org.xquark.schema.Particle;
import org.xquark.schema.Schema;
import org.xquark.schema.SchemaConstants;
import org.xquark.schema.SchemaManager;
import org.xquark.schema.SequenceModelGroup;
import org.xquark.schema.SimpleType;
import org.xquark.schema.Type;
import org.xquark.xquery.parser.XQueryException;
import org.xquark.xquery.parser.XQueryExpression;

public class BuildSchemaTypeVisitor extends DefaultQTypeVisitor {

	private Type resultType = null;

	private QName resultQName = null;

	private Schema schema = null;

	private SchemaManager schemamanager = null;

	private static final String XML_SCHEMA = "http://www.w3.org/2001/XMLSchema";

	public BuildSchemaTypeVisitor() throws TypeException {
		this.schemamanager = new SchemaManager();
		this.schema = new Schema("http://www.xquark.org/TRANSFORMATION", null, schemamanager);
		try {
			schemamanager.putSchema(schema);
		} catch (SAXException saxe) {
			throw new TypeException("Could not add schema to manager");
		}
	}

	public Type getResultType() {
		if (resultType != null)
			return resultType;
		return getType(ElementTools.SC_ANYTYPE);
	}

	public QName getResultQName() {
		return resultQName;
	}

	private Type getType(String typeStr) {
		return schemamanager.getType(XML_SCHEMA, typeStr);
	}

	public void visit(QTypeAtom arg) throws XQueryException {
		SimpleType type = (SimpleType) arg.getType();
		resultType = type;
		resultQName = null;
	}

	public void visit(QTypeAttribute arg) throws XQueryException {
		ComplexType complexType = new ComplexType(this.schema, null, null);
		complexType.setBaseType((SimpleType) getType(ElementTools.SC_STRING));
		complexType.setExtension(true);
		complexType.setContentModel(new ContentModel((SimpleType) getType(ElementTools.SC_STRING)));
		resultType = complexType;
		resultQName = null;
	}

	public void visit(QTypeComment arg) throws XQueryException {
		// do nothing
		resultType = null;
		resultQName = null;
	}

	public void visit(QTypeDocument arg) throws XQueryException {
		// do nothing
		resultType = null;
		resultQName = null;
	}

	public void visit(QTypeElement arg) throws XQueryException {
		resultType = arg.getType();
		if (arg.getName() instanceof org.xquark.xquery.parser.QName) {
			org.xquark.xquery.parser.QName qname = (org.xquark.xquery.parser.QName) arg.getName();
			resultQName = new QName(qname.getNameSpace(), qname.getLocalName());
		} else
			resultQName = null;
	}

	public void visit(QTypeProcessingInstruction arg) throws XQueryException {
		// do nothing
		resultType = null;
		resultQName = null;
	}

	public void visit(QTypeSequence arg) throws XQueryException {
		List list = arg.getList();
		if (list == null || list.isEmpty()) {
			resultType = null;
			resultQName = null;
			return;
		}
		int size = list.size();
		ElementTools eTools = new ElementTools(size, arg.getName(), schema, schemamanager);
		ComplexType complexType = new ComplexType(this.schema, null, null);
		SequenceModelGroup sequenceModel = new SequenceModelGroup();
		for (int i = 0; i < size; i++) {
			QType stype = (QType) list.get(i);
			int subclass = stype.getSubClass();
			if (subclass == QType.ATOM) {
				eTools.buildQTypeAtom(stype);
			} else if (subclass == QType.TEXT) {
				eTools.buildQTypeText(stype);
			} else if (subclass == QType.ATTRIBUTE) {
				eTools.buildQTypeAttribute((QTypeAttribute) stype, complexType);
			} else if (subclass == QType.ELEMENT) {
				eTools.buildQTypeElement(stype, complexType, sequenceModel, stype.getMinOccurence() == 1, stype
						.isMultiple());
			} else if (subclass == QType.UNION) {
				eTools.buildQTypeUnion(stype, complexType, sequenceModel);
			} else {
				// ignore other types
				resultType = null;
				resultQName = null;
				return;
			}
		}
		if ((eTools.getCountAttribute() == eTools.getSize() - eTools.getCountSimpleType()) && (eTools.getMixed())) {
			complexType.setBaseType((SimpleType) getType(ElementTools.SC_STRING));
			complexType.setExtension(true);
			complexType.setContentModel(new ContentModel((SimpleType) getType(ElementTools.SC_STRING)));
		}
		// case of <tag (attr)* > [simpleType] </tag>, like <a b="1"> 23 </a>
		else if ((eTools.getCountAttribute() == eTools.getSize() - 1) && (eTools.getMixed())) {
			complexType.setBaseType(eTools.getSimpleType());
			complexType.setExtension(true);
			complexType.setContentModel(new ContentModel(eTools.getSimpleType()));

		}
		// all others cases...
		else {
			ContentModel contentModel = new ContentModel();
			contentModel.setModel(new Particle(sequenceModel));
			if (eTools.getMixed())
				contentModel.setContentType(SchemaConstants.MIXED);
			else
				contentModel.setContentType(SchemaConstants.ELEMENT_ONLY);
			complexType.setContentModel(contentModel);
		}
		resultType = complexType;
		resultQName = null;
	}

	// transform QTypeUnion into complex type with choice model group

	public void visit(QTypeUnion arg) throws XQueryException {

		List list = arg.getList();
		if (list == null || list.isEmpty()) {
			resultType = null;
			resultQName = null;
			return;
		}
		int size = list.size();
		ElementTools eTools = new ElementTools(size, arg.getName(), schema, schemamanager);
		ComplexType complexType = new ComplexType(this.schema, null, null);
		ChoiceModelGroup choiceModel = new ChoiceModelGroup();
		for (int i = 0; i < size; i++) {
			QType stype = (QType) list.get(i);
			int subclass = stype.getSubClass();
			if (subclass == QType.ATOM) {
				eTools.buildQTypeAtom(stype);
			} else if (subclass == QType.TEXT) {
				eTools.buildQTypeText(stype);
			} else if (subclass == QType.ATTRIBUTE) {
				eTools.buildQTypeAttribute((QTypeAttribute) stype, complexType);
			} else if (subclass == QType.ELEMENT) {
				eTools.buildQTypeElement(stype, complexType, choiceModel, stype.getMinOccurence() == 1, stype
						.isMultiple());
			} else if (subclass == QType.SEQUENCE) {
				SequenceModelGroup sequenceModel = new SequenceModelGroup();
				eTools.buildQTypeSequence(stype, complexType, sequenceModel);
				choiceModel.add(new Particle(sequenceModel));
			} else {
				// ignore other types
				resultType = null;
				resultQName = null;
				return;
			}
		}
		if ((eTools.getCountAttribute() == eTools.getSize() - eTools.getCountSimpleType()) && (eTools.getMixed())) {
			complexType.setBaseType((SimpleType) getType(ElementTools.SC_STRING));
			complexType.setExtension(true);
			complexType.setContentModel(new ContentModel((SimpleType) getType(ElementTools.SC_STRING)));
		}
		// case of <tag (attr)* > [simpleType] </tag>, like <a b="1"> 23 </a>
		else if ((eTools.getCountAttribute() == eTools.getSize() - 1) && (eTools.getMixed())) {
			complexType.setBaseType(eTools.getSimpleType());
			complexType.setExtension(true);
			complexType.setContentModel(new ContentModel(eTools.getSimpleType()));

		}
		// all others cases...
		else {
			ContentModel contentModel = new ContentModel();
			contentModel.setModel(new Particle(choiceModel));
			if (eTools.getMixed())
				contentModel.setContentType(SchemaConstants.MIXED);
			else
				contentModel.setContentType(SchemaConstants.ELEMENT_ONLY);
			complexType.setContentModel(contentModel);
		}
		resultType = complexType;
		resultQName = null;
	}

	public void visit(QTypeText arg) throws XQueryException {
		resultType = arg.getType();
		resultQName = null;
	}

}
